"""
Concrete :class:`~.base.TrackerJobsBase` subclass for MTV
"""

import functools
import re

from ... import errors, jobs, utils
from ...utils.release import ReleaseType
from ..base import TrackerJobsBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class MtvTrackerJobs(TrackerJobsBase):

    @functools.cached_property
    def jobs_before_upload(self):
        # NOTE: Keep in mind that the order of jobs is important for
        #       isolated_jobs: The final job is the overall result, so if
        #       upload_screenshots_job is listed after description_job,
        #       --only-description is going to print the list of uploaded
        #       screenshot URLs.
        return (

            # Interactive jobs
            self.playlists_job,
            self.category_job,
            self.imdb_job,
            self.tmdb_job,
            self.scene_check_job,
            self.title_job,

            # Background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.bdinfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,
            self.description_job,
        )

    @property
    def isolated_jobs(self):
        if self.options.get('only_description', False):
            return self.get_job_and_dependencies(self.description_job)
        elif self.options.get('only_title', False):
            return self.get_job_and_dependencies(
                self.title_job,
            )
        else:
            # Activate all jobs
            return ()

    @functools.cached_property
    def category_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('category'),
            label='Category',
            precondition=self.make_precondition('category_job'),
            autodetect=self.autodetect_category,
            options=[
                (c['label'], c['value'])
                for c in self._categories
            ],
            callbacks={
                'finished': self.handle_category_chosen,
            },
            **self.common_job_args(),
        )

    _categories = (
        {'label': 'HD Season', 'value': '5', 'type': ReleaseType.season},
        {'label': 'HD Episode', 'value': '3', 'type': ReleaseType.episode},
        {'label': 'HD Movie', 'value': '1', 'type': ReleaseType.movie},
        {'label': 'SD Season', 'value': '6', 'type': ReleaseType.season},
        {'label': 'SD Episode', 'value': '4', 'type': ReleaseType.episode},
        {'label': 'SD Movie', 'value': '2', 'type': ReleaseType.movie},
    )

    _category_value_type_map = {
        c['value']: c['type']
        for c in _categories
    }

    def autodetect_category(self, _):
        # "HD" or "SD"
        if utils.mediainfo.video.get_resolution_int(self.content_path) >= 720:
            resolution = 'HD'
        else:
            resolution = 'SD'

        # "Movie", "Episode" or "Season"
        if self.release_name.type is ReleaseType.movie:
            typ = 'Movie'
        elif self.release_name.type is ReleaseType.season:
            typ = 'Season'
        elif self.release_name.type is ReleaseType.episode:
            typ = 'Episode'
        else:
            raise RuntimeError(f'Unsupported type: {self.release_name.type}')

        category = f'{resolution} {typ}'
        _log.debug('Autodetected category: %r', category)
        return category

    @property
    def chosen_release_type(self):
        """
        :class:`~.types.ReleaseType` enum derived from :attr:`category_job` or
        `None` if :attr:`category_job` is not finished yet
        """
        if self.category_job.is_finished:
            choice = self.get_job_attribute(self.category_job, 'choice')
            return self._category_value_type_map.get(choice)

    @functools.cached_property
    def imdb_job(self):
        imdb_job = super().imdb_job
        imdb_job.no_id_ok = True
        return imdb_job

    @functools.cached_property
    def tmdb_job(self):
        tmdb_job = super().tmdb_job
        tmdb_job.no_id_ok = True
        tmdb_job.prejobs += (self.imdb_job,)
        tmdb_job.precondition = self.make_precondition('tmdb_job', precondition=self.no_imdb_id_available)
        return tmdb_job

    def no_imdb_id_available(self):
        return not bool(self.imdb_id)

    def handle_category_chosen(self, category_job_):
        """
        Update the category wherever it matters so the user only has to correct the category
        once
        """
        new_type = self.chosen_release_type
        if new_type is not None:
            old_type = self.imdb_job.query.type
            _log.debug('Updating release type from %s to %s', old_type, new_type)

            # Update title
            if not self.title_job.is_finished:
                self.release_name.type = new_type
            else:
                _log.debug('Not updating title because title job is already finished')

            # Update IMDb query
            if not self.imdb_job.is_finished:
                self.imdb_job.query.type = new_type
            else:
                _log.debug('Not updating IMDb query because IMDb job is already finished')

            # Update TMDb query
            if not self.tmdb_job.is_finished:
                self.tmdb_job.query.type = new_type
            else:
                _log.debug('Not updating TMDb query because TMDb job is already finished')

    release_name_separator = '.'

    release_name_translation = {
        'edition': {
            re.compile(r"^Director's Cut$"): 'DC',
        },
        'group': {
            re.compile(r'^NOGROUP$'): 'NOGRP',
        },
    }

    @functools.cached_property
    def title_job(self):
        """
        :class:`~.jobs.dialog.TextFieldJob` instance with text set to the
        release title

        Unlike :attr:`~.TrackerJobsBase.release_name_job`, this uses the
        original scene release name for movie and episode scene releases.
        """
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('title'),
            label='Title',
            precondition=self.make_precondition('title_job'),
            prejobs=(
                self.category_job,
                self.scene_check_job,
                self.imdb_job,
                self.tmdb_job,
            ),
            text=self.generate_title,
            validator=self.validate_title,
            **self.common_job_args(),
        )

    async def generate_title(self):
        try:
            if (
                    self.chosen_release_type in (ReleaseType.movie, ReleaseType.episode)
                    and self.scene_check_job.is_scene_release
            ):
                # Use the original scene release name instead of a
                # standardized release name. We rely on scene_check_job
                # making sure that the file/directory name is the correct
                # release name.
                search_results = await utils.predbs.MultiPredbApi().search(self.content_path)
                # We can't rely on finding a search result because the user might have marked
                # this as a scene release manually, (e.g. because no predb has indexed that
                # release (yet)).
                if len(search_results) >= 1:
                    return search_results[0]

            # We're dealing with a non-scene release or season pack and may generate a title.
            # First, we try to get more accurate information from a webdb.
            if self.imdb_id:
                _log.debug('Fetching info from IMDb: %r', self.imdb_id)
                await self.release_name.fetch_info(webdb=self.imdb, webdb_id=self.imdb_id)
            elif self.tmdb_id:
                _log.debug('Fetching info from TMDb: %r', self.tmdb_id)
                await self.release_name.fetch_info(webdb=self.tmdb, webdb_id=self.tmdb_id)
            else:
                _log.debug('Not updating autogenerated title')

        except errors.RequestError as e:
            _log.debug('Fetching title failed: %r', e)

        _log.debug('Updated title: %r', str(self.release_name))
        return str(self.release_name)

    def validate_title(self, text):
        if not text.strip():
            raise ValueError('Title must not be empty.')
        super().validate_release_name(text)

    image_host_config = {
        'common': {'thumb_width': 350},
    }

    @functools.cached_property
    def description_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            precondition=self.make_precondition('description_job'),
            prejobs=(
                self.playlists_job,
                self.mediainfo_job,
                self.bdinfo_job,
                self.screenshots_job,
                self.upload_screenshots_job,
            ),
            text=self.generate_description,
            finish_on_success=True,
            read_only=True,
            hidden=True,
            # Don't cache job output because the number of screenshots can be
            # changed by the user between runs.
            **self.common_job_args(ignore_cache=True),
        )

    document_all_videos = False

    async def generate_description(self):
        sections = []
        section = []
        for info in self.video_info.values():
            section.extend(
                f'[mediainfo]{mediainfo}[/mediainfo]\n'
                for mediainfo in info['mediainfos']
            )

            section.extend(
                f'[quote][code]{bdinfo.quick_summary}[/code][/quote]\n'
                for bdinfo in info['bdinfos']
            )

            if info['screenshot_urls']:
                section.append(
                    '[center]'
                    + self.make_screenshots_grid(
                        screenshots=info['screenshot_urls'],
                        columns=2,
                        horizontal_spacer='    ',
                        vertical_spacer='\n\n',
                    )
                    + '[/center]'
                )

            sections.append(''.join(section))
            section.clear()

        description = '\n[hr]\n'.join(sections)
        return (
            description
            + (f'\n\n{self.promotion_bbcode}' if self.promotion_bbcode else '')
        )

    @property
    def post_data_autofill(self):
        return {
            'submit': 'true',
            'MAX_FILE_SIZE': '2097152',
            'fillonly': 'auto fill',
            'category': '0',
            'Resolution': '0',
            'source': '12',
            'origin': '6',
            'title': '',
            'genre_tags': '---',
            'taglist': '',
            'autocomplete_toggle': 'on',
            'image': '',
            'desc': '',
            'fontfont': '-1',
            'fontsize': '-1',
            'groupDesc': '',
            'anonymous': '0',
        }

    @property
    def post_data_upload(self):
        return {
            'submit': 'true',
            'category': self.get_job_attribute(self.category_job, 'choice'),
            'Resolution': '0',
            'source': '12',
            'origin': '6',
            'title': self.get_job_output(self.title_job, slice=0),
            'genre_tags': '---',
            'autocomplete_toggle': 'on',
            'image': '',
            'desc': self.get_job_output(self.description_job, slice=0),
            'fontfont': '-1',
            'fontsize': '-1',
            'groupDesc': self.post_data_group_desc,
            'anonymous': '1' if self.options['anonymous'] else '0',
            'ignoredupes': '1' if self.options['ignore_dupes'] else None,
            # These are probably ignored by the server, but it doesn't cost anything to provide them
            # and maybe eventually someone will pick them up.
            'imdbID': self.imdb_job.selected.get('id', None),
            'tmdbID': self.tmdb_job.selected.get('id', None),
            # 'thetvdbID': ...,
            # 'tvmazeID': ...,
        }

    @property
    def post_data_group_desc(self):
        group_desc = []
        if self.imdb_job.selected:
            group_desc.append(self.imdb_job.selected['url'])
        if self.tmdb_job.selected:
            group_desc.append(self.tmdb_job.selected['url'])

        # Return newline-separated webdb URLs or None.
        if group_desc:
            return '\n'.join(group_desc)



"""
Get information from "BDMV" and "VIDEO_TS" directory trees
"""

from .bluray import is_bluray
from .dvd import is_dvd
from .playlist import Playlist


def is_disc(content_path, *, multidisc=True):
    """Whether `content_path` :func:`~.is_bluray` or :func:`~.is_bluray`"""
    return (
        is_bluray(content_path, multidisc=multidisc)
        or is_dvd(content_path, multidisc=multidisc)
    )




import functools
import textwrap

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import ConditionalContainer, DynamicContainer, HSplit, VSplit, Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.utils import get_cwidth

from ....utils import browser, webdbs
from .. import widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class WebDbSearchJobWidget(JobWidgetBase):

    is_interactive = True
    total_width = 80
    left_column_width = 36
    right_column_width = total_width - left_column_width - 1
    total_height = 20

    def setup(self):
        self._query_exception = None
        self._at_least_one_search_was_done = False
        self._widgets = {
            # First row
            'id': widgets.TextField(width=15, style='class:dialog.search.info'),
            'query': widgets.InputField(
                style='class:dialog.search.query',
                text=str(self.job.query),
                on_accepted=self.handle_query,
            ),

            # Right column
            'search_results': _SearchResults(width=self.right_column_width),
            'title_original': widgets.TextField(
                style='class:dialog.search.info',
                width=self.right_column_width,
                height=1,
            ),
            'title_english': widgets.TextField(
                style='class:dialog.search.info',
                width=self.right_column_width,
                height=1,
            ),

            # Left column
            'summary': widgets.TextField(
                style='class:dialog.search.info',
                width=self.left_column_width,
                height=8,
            ),
            'genres': widgets.TextField(
                style='class:dialog.search.info',
                width=self.left_column_width,
                height=2,
            ),
            'directors': widgets.TextField(
                style='class:dialog.search.info',
                width=self.left_column_width,
                height=1,
            ),
            'cast': widgets.TextField(
                style='class:dialog.search.info',
                width=self.left_column_width,
                height=2,
            ),
            'countries': widgets.TextField(
                style='class:dialog.search.info',
                width=self.left_column_width,
                height=1,
            ),

            # Poster (rightmost column)
            'poster': widgets.Image(height=self.total_height),
        }

        self.set_query_text(self.job.query)

        self.job.signal.register('search_results', self.handle_search_results)
        self.job.signal.register('searching_status', self.handle_searching_status)
        self.job.signal.register('info_updating', self.handle_info_updating)
        self.job.signal.register('info_updated', self.handle_info_updated)
        self.job.signal.register('query_updated', self.handle_query_updated)

    def invalidate(self, *, warnings=False):
        super().invalidate()
        if warnings:
            try:
                # Invalidate cached warnings
                del self.warnings
            except AttributeError:
                # No warnings have been generated yet
                pass

    def handle_query(self, buffer):
        query_text = self._widgets['query'].text
        if query_text != self._old_query_text:
            try:
                self.job.search(webdbs.Query.from_string(query_text))
            except ValueError as e:
                self._query_exception = e
            else:
                self._query_exception = None
                self.set_query_text(self.job.query)

        else:
            # The same query was accepted twice without changing it.
            # Select focused search result.
            focused = self._widgets['search_results'].focused_result
            if focused is not None:
                self.job.result_selected(focused)
            else:
                self.job.result_selected(None)

            # Query exception is no longer valid
            self._query_exception = None

        # Invalidate warnings to include new self._query_exception
        self.invalidate(warnings=True)

    def handle_searching_status(self, is_searching):
        self._widgets['search_results'].is_searching = is_searching

    def handle_search_results(self, results):
        self._widgets['search_results'].results = results
        self._at_least_one_search_was_done = True
        # Invalidate warnings to display/remove no-result-search-hints
        self.invalidate(warnings=True)

    def handle_info_updating(self, attr):
        self._widgets[attr].is_loading = True
        self.invalidate()

    def handle_info_updated(self, attr, value):
        self._widgets[attr].is_loading = False
        if isinstance(value, (bytes, bytearray)):
            self._widgets[attr].text = value
        elif value:
            self._widgets[attr].text = str(value)
        else:
            self._widgets[attr].text = ''
        self.invalidate()

    def handle_query_updated(self, query):
        self.set_query_text(query)
        for attr in (
                'id',
                'title_original',
                'title_english',
                'summary',
                'genres',
                'directors',
                'cast',
                'countries',
                'poster',
        ):
            self._widgets[attr].text = ''
        self.invalidate()

    def set_query_text(self, text):
        self._widgets['query'].set_text(str(text), preserve_cursor_position=True)
        self._old_query_text = self._widgets['query'].text

    @functools.cached_property
    def warnings(self):
        """Note for the user to find results"""
        warnings = []

        # User entered invalid query
        if self._query_exception:
            warnings.append(str(self._query_exception))

        # No search results - provide some hints
        if (
                # Don't show hints if we are still starting up and the first
                # search is not yet happening.
                self._at_least_one_search_was_done
                # Don't show hints if we are currently searching.
                and not self.job.is_searching
                # Don't show hints if there are search results.
                and not self._widgets['search_results'].results
        ):
            # Generic hints
            warnings.append(
                textwrap.fill(
                    'Try removing or adjusting "year:…" or "type:…". '
                    'You can provide an exact ID by searching for "id:…".',
                    width=self.total_width,
                ),
            )

            # Search hints for this specific webdb
            if self.job.db.no_results_info:
                warnings.append(
                    textwrap.fill(
                        self.job.db.no_results_info,
                        width=self.total_width,
                    )
                )

            # Maybe provide no ID as a valid job result
            if self.job.no_id_ok:
                warnings.append(
                    textwrap.fill(
                        (
                            f'Press Enter to provide no {self.job.db.label} ID if you are sure '
                            f'"{self.job.query.title}" does not exist on {self.job.db.label}.'
                        ),
                        width=self.total_width,
                    )
                )

        return '\n\n'.join(warnings)

    @functools.cached_property
    def runtime_widget(self):
        w = self._widgets

        # Everything except for the poster on the very right.
        textfields = HSplit(
            children=[
                VSplit([
                    w['query'],
                    widgets.hspacer,
                    widgets.HLabel(
                        text='ID',
                        content=w['id'],
                        style='class:dialog.search.label',
                    ),
                ]),
                ConditionalContainer(
                    filter=Condition(lambda: self.warnings),
                    content=HSplit([
                        widgets.vspacer,
                        Window(
                            FormattedTextControl(
                                lambda: self.warnings,
                                style='class:warning',
                            ),
                        )
                    ]),
                ),
                VSplit([
                    HSplit([
                        widgets.VLabel('Results', w['search_results'], style='class:dialog.search.label'),
                        widgets.VLabel('Original Title', w['title_original'], style='class:dialog.search.label'),
                        widgets.VLabel('Also Known As', w['title_english'], style='class:dialog.search.label'),
                    ]),
                    widgets.hspacer,
                    HSplit([
                        widgets.VLabel('Summary', w['summary'], style='class:dialog.search.label'),
                        widgets.VLabel('Genres', w['genres'], style='class:dialog.search.label'),
                        widgets.VLabel('Director', w['directors'], style='class:dialog.search.label'),
                        widgets.VLabel('Cast', w['cast'], style='class:dialog.search.label'),
                        widgets.VLabel('Country', w['countries'], style='class:dialog.search.label'),
                    ]),
                ]),
            ],
            style='class:dialog.search',
        )

        # `textfields` and the poster.
        dialog = VSplit(
            children=[
                textfields,
                widgets.hspacer,
                w['poster'],
            ],
        )
        # Add a spacer below the whole form to get some breathing room.
        layout = HSplit(
            children=[
                dialog,
                widgets.hspacer,
            ],
        )

        @self.keybindings_local.add('down')
        @self.keybindings_local.add('c-n')
        @self.keybindings_local.add('tab')
        def _(_event):
            prev_focused = self._widgets['search_results'].focused_result
            self._widgets['search_results'].focus_next()
            now_focused = self._widgets['search_results'].focused_result
            if prev_focused != now_focused:
                self.job.result_focused(now_focused)

        @self.keybindings_local.add('up')
        @self.keybindings_local.add('c-p')
        @self.keybindings_local.add('s-tab')
        def _(_event):
            prev_focused = self._widgets['search_results'].focused_result
            self._widgets['search_results'].focus_previous()
            now_focused = self._widgets['search_results'].focused_result
            if prev_focused != now_focused:
                self.job.result_focused(now_focused)

        # Alt-Enter
        @self.keybindings_local.add('escape', 'enter')
        def _(_event):
            url = self._widgets['search_results'].focused_result.url
            browser.open(url)

        return layout


class _SearchResults(DynamicContainer):
    def __init__(self, results=(), width=40):
        self.results = results
        self._year_width = 4
        self._type_width = 6
        self._title_width = width - self._year_width - self._type_width - 2
        self._activity_indicator = widgets.ActivityIndicator()
        super().__init__(
            lambda: Window(
                content=FormattedTextControl(self._get_text_fragments, focusable=False),
                width=width,
                height=14,
                style='class:dialog.search.results',
            )
        )

    @property
    def is_searching(self):
        return self._activity_indicator.active

    @is_searching.setter
    def is_searching(self, value):
        self._activity_indicator.active = bool(value)

    @property
    def results(self):
        return self._results

    @results.setter
    def results(self, results):
        self._results = tuple(results)
        self._focused_index = 0

    @property
    def focused_result(self):
        if self._results:
            return self._results[self._focused_index]
        else:
            return None

    def focus_next(self):
        if self._focused_index < len(self._results) - 1:
            self._focused_index += 1

    def focus_previous(self):
        if self._focused_index > 0:
            self._focused_index -= 1

    def focus_first(self):
        self._focused_index = 0

    def focus_last(self):
        self._focused_index = len(self._results) - 1

    def _get_text_fragments(self):
        if self.is_searching:
            return [('class:dialog.search.results', self._activity_indicator.text)]
        elif not self._results:
            return 'No results'

        frags = []
        for i, result in enumerate(self._results):
            if i == self._focused_index:
                title_style = 'class:dialog.search.results.focused'
                frags.append(('[SetCursorPosition]', ''))
                self._focused_result = result
            else:
                title_style = 'class:dialog.search.results'

            if get_cwidth(result.title) > self._title_width:
                title = result.title[:self._title_width - 1] + '…'
            else:
                title = result.title
            frags.append((title_style, title.ljust(self._title_width)))

            frags.append(('', (
                ' '
                f'{str(result.year or "").rjust(4)}'
                ' '
                f'{str(result.type).rjust(6)}'
            )))

            frags.append(('', '\n'))
        frags.pop()  # Remove last newline
        return frags



"""
Interactive text user interface and job manager
"""

import asyncio
import collections
import types

from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import Layout
from prompt_toolkit.layout.containers import HSplit, Window, to_container
from prompt_toolkit.output import create_output

from . import jobwidgets, style

import logging  # isort:skip
_log = logging.getLogger(__name__)


class TUI:
    def __init__(self):
        # Map JobBase.name to SimpleNamespace with attributes:
        #   job       - JobBase instance
        #   widget    - JobWidgetBase instance
        #   container - prompttoolkit.Container instance
        self._jobs = collections.defaultdict(lambda: types.SimpleNamespace())
        self._focused_jobinfo = None
        self._app = self._make_app()
        self._exception = None
        self._run_task = None
        self._run_task_cancelled = False
        self._loop = asyncio.get_event_loop()
        self._loop.set_exception_handler(self._handle_exception)

    def _handle_exception(self, loop, context):
        exception = context.get('exception')
        if exception:
            _log.debug('Caught unhandled exception: %r', exception)
            _log.debug('Unhandled exception context: %r', context)
            if not self._exception:
                self._exception = exception
            self._exit()

    def _make_app(self):
        self._jobs_container = HSplit(
            # FIXME: Layout does not accept an empty list of children, so we add
            #        an empty Window that doesn't display anything.
            #        https://github.com/prompt-toolkit/python-prompt-toolkit/issues/1257
            children=[Window()],
            style='class:default',
        )
        self._layout = Layout(self._jobs_container)

        kb = KeyBindings()

        @kb.add('escape')
        @kb.add('c-g')
        @kb.add('c-q')
        @kb.add('c-c')
        def _(_event, self=self):
            _log.debug('Terminating all jobs because the user terminated the application')
            self._terminate_jobs()
            self._exit()

        @kb.add('escape', 'I')
        def _(_event, self=self):
            _log.debug('=== CURRENT JOBS ===')
            for jobinfo in self._jobs.values():
                job = jobinfo.job
                if job.is_finished:
                    state = '[finished]'
                elif job.was_started:
                    state = '[running ]'
                else:
                    state = '[disabled]'
                _log.debug(' %s %s (%d tasks):', state, job.name, len(job._tasks))
                for task in job._tasks:
                    _log.debug('   %r', task)

            _log.debug('Focused widget: %r', self._layout.current_control)
            _log.debug('Focused jobinfo: %r', self._focused_jobinfo)

        app = Application(
            # Write TUI to stderr if stdout is redirected. This is useful for
            # allowing the user to make decisions in the TUI (e.g. selecting an
            # item from search results) while redirecting the final output
            # (e.g. an IMDb ID).
            output=create_output(always_prefer_tty=True),
            layout=self._layout,
            key_bindings=kb,
            style=style.style,
            full_screen=False,
            erase_when_done=False,
            mouse_support=False,
            # Determine the currently active job *after* Application was invalidated. At the time of
            # this writing, this is important so we don't try to focus widgets from a job that has
            # just finished. This may result in RuntimeErrors because a signal is emitted after it's
            # job is finished.
            before_render=self._update_jobs_container,
        )
        return app

    def add_jobs(self, *jobs):
        """Add :class:`~.jobs.base.JobBase` instances"""
        for job in jobs:
            self._add_job(job)

        # Add job widgets to the main container widget
        self._update_jobs_container()

        # Register signal callbacks
        self._connect_jobs(jobs)

    def _add_job(self, job):
        if job.name in self._jobs:
            if job is not self._jobs[job.name].job:
                raise RuntimeError(f'Conflicting job name: {job.name}')
        else:
            self._jobs[job.name].job = job
            self._jobs[job.name].widget = jobwidgets.JobWidget(job, self._app)
            self._jobs[job.name].container = to_container(self._jobs[job.name].widget)

    # We accept one argument because the on_invalidate callback passes the
    # Application instance
    def _update_jobs_container(self, _=None):
        enabled_jobs = self._enabled_jobs

        # Unfocus focused job if it is finished.
        if self._focused_jobinfo and self._focused_jobinfo.job.is_finished:
            # _log.debug('UNFOCUSING: %r', self._focused_jobinfo)
            self._focused_jobinfo = None

        # Don't change focus if we already have a focused job. If another job
        # becomes interactive asynchronously (e.g. because a background job
        # finished), it must not steal focus from the currently focused job.
        if not self._focused_jobinfo:
            # Focus next interactive job.
            for jobinfo in enabled_jobs:
                if (
                        jobinfo.widget.is_interactive
                        and jobinfo.job.was_started
                        and not jobinfo.job.is_finished
                ):
                    # _log.debug('FOCUSING NEXT INTERACTIVE JOB: %s', jobinfo.job.name)
                    self._focused_jobinfo = jobinfo
                    break
        #     else:
        #         _log.debug('NO FOCUSABLE WIDGET FOUND: %r', [
        #             {
        #                 'name': jobinfo.job.name,
        #                 'is_interactive': jobinfo.widget.is_interactive,
        #                 'was_started': jobinfo.job.was_started,
        #                 'is_finished': jobinfo.job.is_finished,
        #             }
        #             for jobinfo in enabled_jobs
        #         ])
        # else:
        #     _log.debug('PRESERVING FOCUS: %r', self._focused_jobinfo.job.name)

        # Display all background jobs, finished jobs and the first unfinished
        # interactive job.
        self._jobs_container.children[:] = (
            jobinfo.container
            for jobinfo in enabled_jobs
            if (
                jobinfo is self._focused_jobinfo
                or jobinfo.job.is_finished
                or not jobinfo.widget.is_interactive
            )
        )

        if self._focused_jobinfo:
            # Actually focus the focused job.
            try:
                self._layout.focus(self._focused_jobinfo.container)
            except ValueError:
                # A job may hardcode `is_interactive = True` even though it
                # currently is not focusable. This happens if the job is still
                # autodetecting before allowing the user to fix/confirm.
                pass

    def _connect_jobs(self, jobs):
        for job in jobs:
            # Every time a job finishes, other jobs can become enabled due to
            # the dependencies on other jobs or other conditions. We also want
            # to display the next interactive job when an interactive job is
            # done.
            job.signal.register('finished', self._handle_job_finished)

            # A job can also signal explicitly that we should update the job
            # widgets, e.g. to start previously disabled jobs.
            job.signal.register('refresh_ui', self._refresh_jobs)

    def _handle_job_finished(self, finished_job):
        assert finished_job.is_finished, f'{finished_job.name} is actually not finished'

        # Start and/or display the next interactive job. This also generates the
        # regular output if all output was read from cache and the TUI exits
        # immediately.
        self._refresh_jobs()

        # Terminate all jobs and exit if job finished with non-zero exit code
        if finished_job.exit_code != 0:
            _log.debug('Terminating all jobs because job failed: %s: exit_code=%r',
                       finished_job.name, finished_job.exit_code)
            self._terminate_jobs()
            self._exit()

        elif self._all_jobs_finished:
            # Exit application if all jobs finished
            _log.debug('All jobs finished')
            self._exit()

    def _refresh_jobs(self):
        self._start_enabled_jobs()
        self._update_jobs_container()
        self._app.invalidate()

    def _start_enabled_jobs(self):
        for jobinfo in self._enabled_jobs:
            job = jobinfo.job
            if not job.was_started and job.autostart:
                job.start()

    @property
    def _enabled_jobs(self):
        return tuple(
            jobinfo
            for jobinfo in self._jobs.values()
            if jobinfo.job.is_enabled
        )

    @property
    def _all_jobs_finished(self):
        enabled_jobs = [jobinfo.job for jobinfo in self._enabled_jobs]
        return all(job.is_finished for job in enabled_jobs)

    def run(self, jobs):
        """
        Block while running `jobs`

        :param jobs: Iterable of :class:`~.jobs.base.JobBase` instances

        :raise: Any exception that occured while running jobs

        :return: :attr:`~.JobBase.exit_code` from the first failed job or 0 for
            success
        """
        self.add_jobs(*jobs)

        # Block until _exit() is called
        self._loop.run_until_complete(self._run())

        # Raise exception or return exit code
        exception = self._get_exception()
        if exception:
            _log.debug('Exception from jobs: %r', exception)
            raise exception
        else:
            for jobinfo in self._enabled_jobs:
                _log.debug('Exit code of %r: %r', jobinfo.job.name, jobinfo.job.exit_code)

            # First non-zero exit_code is the application exit_code
            for jobinfo in self._enabled_jobs:
                if jobinfo.job.exit_code != 0:
                    _log.debug('Exiting with exit code from %s: %r', jobinfo.job.name, jobinfo.job.exit_code)
                    return jobinfo.job.exit_code
            return 0

    async def _run(self):
        # Call each job's run() method or read cache. This must be done
        # asynchronously because we need a running asyncio event loop to make
        # JobBase.add_task() work.
        self._start_enabled_jobs()

        # Run the TUI in a task that can be cancelled. This avoids calling
        # self._app.exit(), which doesn't force self._app.run_async() to return.
        self._run_task = self._loop.create_task(
            self._app.run_async(set_exception_handler=False)
        )
        try:
            await self._run_task
        except asyncio.CancelledError:
            _log.debug('Application was cancelled: %r', self._run_task)
            self._terminate_jobs()

        # Wait for all jobs. If they were terminated, they should also be able
        # to clean up their tasks and whatever else there is to do.
        enabled_jobs = [jobinfo.job for jobinfo in self._enabled_jobs]
        for job in enabled_jobs:
            if job.was_started and not job.is_finished:
                await job.wait_finished()

    def _exit(self):
        if self._run_task_cancelled:
            _log.debug('Application already exited: %r', self._run_task)
        elif self._run_task:
            _log.debug('Exiting application: %r', self._run_task)
            self._run_task.cancel()
            self._run_task_cancelled = True
        else:
            # We end up here if _exit() is called before the
            # self._app.run_async() task is created. If we call call_soon()
            # instead of call_later(), the TUI is not printed (which we
            # want), probably because self._app.run_async() is cancelled
            # before it can do that.
            _log.debug('Trying to exit the application again soon')
            self._loop.call_later(0, self._exit)

    def _terminate_jobs(self):
        for jobinfo in self._jobs.values():
            job = jobinfo.job
            if not job.is_finished:
                _log.debug('Terminating %s', job.name)
                job.terminate()

    def _get_exception(self):
        # Return uncaught exception, which should be caught by
        # self._handle_exception().
        if self._exception:
            return self._exception

        # Return first exception from any job or None
        for jobinfo in self._jobs.values():
            job = jobinfo.job
            if job.raised:
                _log.debug('Exception from %s: %r', job.name, job.raised)
                return job.raised



"""
Dump frames from video file
"""

import os

from .. import errors, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


def _ffmpeg_executable():
    if utils.os_family() == 'windows':
        return 'ffmpeg.exe'
    else:
        return 'ffmpeg'


def _make_screenshot_cmd(video_file, timestamp, screenshot_file):
    # ffmpeg's "image2" image file muxer uses "%" for string formatting, so we
    # must escape "%" in `video_file`
    screenshot_file = screenshot_file.replace('%', '%%')

    # -vf argument from:
    # https://rendezvois.github.io/video/screenshots/programs-choices/#ffmpeg

    vf = {}

    # Some magic to prevent grey screenshots for some BDMV releases. (Maybe related to VC-1?)
    if (
            utils.disc.is_bluray(video_file)
            or utils.fs.file_extension(video_file).lower() == 'm2ts'
    ):
        vf['fps'] = ('1/60',)

    # Always fix aspect ratio in case video is anamorphic. Doesn't matter if it isn't.
    vf['scale'] = ["'max(sar,1)*iw'", "'max(1/sar,1)*ih'"]

    if utils.mediainfo.video.is_bt2020(video_file):
        vf['scale'].extend((
            'in_h_chr_pos=0',
            'in_v_chr_pos=0',
            'in_color_matrix=bt2020',
        ))

    elif utils.mediainfo.video.is_bt709(video_file):
        vf['scale'].extend((
            'in_h_chr_pos=0',
            'in_v_chr_pos=128',
            'in_color_matrix=bt709',
        ))

    elif utils.mediainfo.video.is_bt601(video_file):
        vf['scale'].extend((
            'in_h_chr_pos=0',
            'in_v_chr_pos=128',
            'in_color_matrix=bt601',
        ))

    vf['scale'].append('flags=' + '+'.join((
        'full_chroma_int',
        'full_chroma_inp',
        'accurate_rnd',
        'spline',
    )))

    vf = ','.join(
        f'{filtername}=' + ':'.join(filtervalue)
        for filtername, filtervalue in vf.items()
    )

    return (
        _ffmpeg_executable(),
        '-y',
        '-hide_banner',
        '-loglevel', 'error',
        '-ss', str(timestamp),
        '-i', f'file:{video_file}',
        '-vf', vf,
        '-pix_fmt', 'rgb24',
        '-vframes', '1',
        f'file:{screenshot_file}',
    )


def screenshot(video_file, timestamp, screenshot_file):
    """
    Create single screenshot from video file

    :param str video_file: Path to video file
    :param timestamp: Time location in the video
    :type timestamp: :class:`~.Timestamp` (or any :class:`int` or :class:`float`)
    :param str screenshot_file: Path to screenshot file

    .. note:: It is important to use the returned file path because it is passed
              through :func:`~.fs.sanitize_path` to make sure it can exist.

    :raise ScreenshotError: if something goes wrong

    :return: Path to screenshot file
    """
    # See if `video_file` is readable before we do further checks and launch ffmpeg, which will
    # produce a long and unexpected error message.
    try:
        utils.fs.assert_file_readable(video_file)
    except errors.ContentError as e:
        raise errors.ScreenshotError(e) from e

    # Validate timestamp.
    if not isinstance(timestamp, utils.types.Timestamp):
        try:
            timestamp = utils.types.Timestamp(timestamp)
        except (TypeError, ValueError) as e:
            raise errors.ScreenshotError(f'Invalid timestamp: {timestamp!r}') from e

    # Make `screenshot_file` compatible to the file system.
    screenshot_file = utils.fs.sanitize_path(screenshot_file)

    # Ensure timestamp is within range.
    duration = utils.types.Timestamp(utils.mediainfo.get_duration(video_file))
    if timestamp > duration:
        raise errors.ScreenshotError(f'Timestamp is too close to or after end of video ({duration}): {timestamp}')

    # Make screenshot.
    cmd = _make_screenshot_cmd(video_file, timestamp, screenshot_file)
    output = utils.subproc.run(cmd, ignore_errors=True, join_stderr=True)
    if not os.path.exists(screenshot_file):
        import shlex
        raise errors.ScreenshotError(
            f'{video_file}: Failed to create screenshot at {timestamp}: {output}\n'
            + ' '.join(shlex.quote(arg) for arg in cmd)
        )
    else:
        return screenshot_file


def _make_resize_cmd(image_file, dimensions, resized_file):
    # ffmpeg's "image2" image file muxer uses "%" for string formatting
    resized_file = resized_file.replace('%', '%%')
    return (
        _ffmpeg_executable(),
        '-y',
        '-hide_banner',
        '-loglevel', 'error',
        '-i', f'file:{image_file}',
        '-vf', f'scale={dimensions}:force_original_aspect_ratio=decrease',
        f'file:{resized_file}',
    )


def resize(image_file, *, width=0, height=0, target_directory=None, target_filename=None, overwrite=False):
    """
    Resize image, preserve aspect ratio

    :param image_file: Path to source image
    :param width: Desired image width in pixels or `0`
    :param height: Desired image height in pixels or `0`
    :param target_directory: Where to put the resized image or `None` to use the
        parent directory of `image_file`
    :param target_filename: File name of resized image or `None` to generate a
        name from `image_file`, `width` and `height`
    :param bool overwrite: Whether to overwrite the resized image file if it
        already exists

    If `width` and `height` are falsy (the default) return `image_file` if
    `target_directory` and `target_filename` are falsy or copy `image_file` to
    the target path.

    .. note:: It is important to use the returned file path because it is passed
              through :func:`~.fs.sanitize_path` to make sure it can exist.

    :raise ImageResizeError: if resizing fails

    :return: Path to resized or copied image
    """
    try:
        utils.fs.assert_file_readable(image_file)
    except errors.ContentError as e:
        raise errors.ImageResizeError(e) from e

    if width and width < 1:
        raise errors.ImageResizeError(f'Width must be greater than zero: {width}')
    elif height and height < 1:
        raise errors.ImageResizeError(f'Height must be greater than zero: {height}')

    dimensions_map = {'width': int(width), 'height': int(height)}
    ext_args = {'minlen': 3, 'maxlen': 4}

    def get_target_filename():
        if target_filename:
            filename = utils.fs.strip_extension(target_filename, **ext_args)
            extension = utils.fs.file_extension(target_filename, **ext_args)
            if not extension:
                extension = utils.fs.file_extension(image_file, **ext_args)
        else:
            filename = utils.fs.basename(utils.fs.strip_extension(image_file, **ext_args))
            dimensions = ','.join(f'{k}={v}' for k, v in dimensions_map.items() if v)
            if dimensions:
                filename += f'.{dimensions}'
            extension = utils.fs.file_extension(image_file, **ext_args)

        if extension:
            filename += f'.{extension}'
        else:
            filename += '.jpg'

        return filename

    def get_target_directory():
        if target_directory:
            return str(target_directory)
        else:
            return utils.fs.dirname(image_file)

    # Assemble full target filepath and make sure it can exist
    target_filepath = utils.fs.sanitize_path(
        os.path.join(get_target_directory(), get_target_filename()),
    )

    if not overwrite and os.path.exists(target_filepath):
        _log.debug('Already resized: %r', target_filepath)
        return target_filepath

    if not width and not height:
        # Nothing to resize
        if target_filepath != str(image_file):
            # Copy image_file to target_filepath
            try:
                utils.fs.mkdir(utils.fs.dirname(target_filepath))
            except errors.ContentError as e:
                raise errors.ImageResizeError(e) from e

            import shutil
            try:
                return str(shutil.copy2(image_file, target_filepath))
            except OSError as e:
                msg = e.strerror if e.strerror else str(e)
                raise errors.ImageResizeError(
                    f'Failed to copy {image_file} to {target_filepath}: {msg}'
                ) from e
        else:
            # Nothing to copy
            return str(image_file)

    ffmpeg_params = ':'.join(
        f'{k[0]}={v if v else -1}'
        for k, v in dimensions_map.items()
    )
    cmd = _make_resize_cmd(image_file, ffmpeg_params, target_filepath)
    output = utils.subproc.run(cmd, ignore_errors=True, join_stderr=True)
    if not os.path.exists(target_filepath):
        error = output or 'Unknown reason'
        raise errors.ImageResizeError(f'Failed to resize: {error}')
    else:
        return str(target_filepath)


# NOTE: Most of the optimization is achieved at level 1 with ~40 % smaller
#       files. Anything higher seems to only reduce by tens of kB or less.
_optimization_levels = {
    'low': '1',
    'medium': '2',
    'high': '4',
    'placebo': 'max',
}

optimization_levels = (*tuple(_optimization_levels), 'none', 'default')
"""Valid `level` arguments for :func:`optimize`"""


def optimize(image_file, output_file=None, level=None):
    """
    Optimize PNG image size

    :path image_file: Path to PNG File
    :path output_file: Path to optimized `image_file` or any falsy value to
        overwrite `image_file`
    :path str,int level: Optimiziation level (``"low"``, ``"medium"``,
        ``"high"``) or ``"default"`` to use recommended level or ``"none"`` /
        `None` to not do any optimization

    If the optimization fails and `image_file` does not end with ".png", it is assumed that it is
    not a PNG and the original file is returned unoptimized.

    :return: path to optimized PNG file
    :raise ImageOptimizeError: if the optimization fails
    """
    if level not in ('none', None):
        if level == 'default':
            level = 'medium'

        try:
            opt = _optimization_levels[str(level)]
        except KeyError as e:
            raise errors.ImageOptimizeError(f'Invalid optimization level: {level}') from e

        cmd = [
            'oxipng', '--preserve',
            '--opt', opt,
            '--interlace', '0',  # Remove any interlacing
            '--strip', 'safe',   # Remove irrelevant metadata
            str(image_file),
        ]

        if output_file:
            sanitized_output_file = utils.fs.sanitize_path(output_file)
            cmd.extend(('--out', sanitized_output_file))
            return_value = sanitized_output_file
        else:
            return_value = image_file

        # oxipng prints errors AND info messages to stderr, so we can't use output on stderr as an
        # indicator of failure. Instead, we check if the output file exists
        error_message, exitcode = utils.subproc.run(cmd, join_stderr=True, return_exitcode=True)
        error_message = error_message.strip()
        if exitcode == 0:
            return return_value
        elif utils.fs.file_extension(image_file).lower() == 'png':
            if error_message:
                raise errors.ImageOptimizeError(f'Failed to optimize: {error_message}')
            else:
                raise errors.ImageOptimizeError('Failed to optimize for unknown reason')
        else:
            # This is probably not a PNG and we can just fail silently and return the
            # unoptimized original.
            return image_file



"""
Get information from "BDMV" directory trees
"""

import os

import pyparsebluray

from ... import utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


def is_bluray(content_path, *, multidisc=False):
    """
    Whether `content_path` contains a "BDMV" subdirectory

    If `multidisc` is truthy, also look for a "BDMV" directory in any subdirectory, but not
    recursively.
    """
    if os.path.isdir(content_path):
        if os.path.isdir(os.path.join(content_path, 'BDMV')):
            return True

        if multidisc:
            for subdir in utils.fs.listdir(content_path):
                if os.path.isdir(os.path.join(content_path, subdir, 'BDMV')):
                    return True

    return False


def get_disc_paths(content_path):
    """
    Return sequence of directory paths beneath `content_path` that contain a "BDMV" directory
    """

    discpaths = []

    # Most likely single-disc release.
    if os.path.isdir(os.path.join(content_path, 'BDMV')):
        discpaths.append(content_path)

    # Find more discs in subdirectories.
    if os.path.isdir(content_path):
        for name in utils.fs.listdir(content_path):
            subpath = os.path.join(content_path, name)
            if is_bluray(subpath):
                discpaths.append(subpath)

    return tuple(discpaths)


def get_playlists(discpath):
    """
    Return sequence of :class:`~.Playlist` instances from BDMV subdirectory

    Return empty sequence if no playlists are found.

    Playlists with a runtime of less than 3 minutes are ignored.

    Each playlist's :attr:`~.Playlist.filepath` is the .MPLS file and the :attr:`~.Playlist.items`
    are .M2TS files.

    :param discpath: Path to directory that contains a "BDMV" subdirectory
    """
    playlist_directory = os.path.join(discpath, 'BDMV', 'PLAYLIST')
    stream_directory = os.path.join(discpath, 'BDMV', 'STREAM')
    if os.path.isdir(playlist_directory) and os.path.isdir(stream_directory):
        # Create Mpls instances and filter out any garbage.
        mplss = _filter_mplss(
            Mpls(mpls_filepath)
            for mpls_filepath in utils.fs.file_list(playlist_directory, extensions=('mpls',))
        )
    else:
        mplss = ()

    def get_duration(mpls):
        # Combined runtime for all playlist items.
        # "Times are expressed in 45 KHz, so you'll need to divide by 45000 to get them in seconds."
        # https://github.com/lw/BluRay/wiki/PlayItem#intime-and-outtime
        duration = 0
        for item in mpls['playlist']['play_items']:
            duration += (item['outtime'] - item['intime']) / 45000
        return duration

    # Create Playlist instances from Mpls instances and filter out any garbage.
    playlists = _filter_playlists(
        utils.disc.Playlist(
            discpath=discpath,
            # Playlist file (.mpls).
            filepath=mpls.filepath,
            # Video file(s) (.m2ts).
            items=tuple(
                os.path.join(
                    stream_directory,
                    item['clip_information_filename'] + '.m2ts',
                )
                for item in mpls['playlist']['play_items']
            ),
            # Combined runtime of all items.
            duration=get_duration(mpls),
        )
        for mpls in mplss
    )

    utils.disc.playlist.mark_main_playlists(playlists)
    return playlists


# This function filters Mpls instances.
def _filter_mplss(mplss):
    def is_empty(mpls):
        # Ignore empty playlists.
        if mpls['playlist']['play_items']:
            return False
        else:
            _log.debug(f'{mpls.filepath}: Mpls is empty: {mpls}')
            return True

    def repeats_clips(mpls):
        # Ignore playlists that repeat the same m2ts.

        def play_item_id(playitem):
            return (playitem['clip_information_filename'], playitem['intime'], playitem['outtime'])

        play_item_ids = tuple(
            play_item_id(play_item)
            for play_item in mpls['playlist']['play_items']
        )
        for play_item_id in play_item_ids:
            if play_item_ids.count(play_item_id) > 2:
                _log.debug(f'{mpls.filepath}: Found repeated play item: {play_item_id}')
                return True
        return False

    return tuple(
        mpls for mpls in mplss
        if (
                not is_empty(mpls)
                and not repeats_clips(mpls)
        )
    )


# This function filters Playlist instances.
def _filter_playlists(playlists):
    def is_too_short(playlist):
        # Ignore playlists with a total runtime of less than 3 minutes.
        if playlist.duration >= 180:
            return False
        else:
            _log.debug(f'{playlist}: Playlist is too short: {playlist.duration}')
            return True

    return tuple(
        playlist for playlist in playlists
        if (
                not is_too_short(playlist)
        )
    )


class Mpls(dict):
    """
    :class:`dict` that reads a Blu-ray ``.mpls`` file

    The provided `filepath` is available as an instance attribute.
    """

    def __init__(self, filepath):
        self.filepath = filepath

        def sanitize(thing):
            if isinstance(thing, pyparsebluray.mpls.movie_playlist.MplsObject):
                thing = vars(thing)

            if isinstance(thing, dict):
                dct = {key: sanitize(value) for key, value in thing.items()}
                del dct['mpls']
                return dct

            elif isinstance(thing, (list, tuple)):
                return tuple(sanitize(value) for value in thing)

            else:
                return thing

        def as_normal_value(mpls_object):
            return sanitize(vars(mpls_object))

        with open(filepath, mode='rb') as fd:
            self['header'] = as_normal_value(pyparsebluray.load_movie_playlist(fd))
            self['appinfo'] = as_normal_value(pyparsebluray.load_app_info_playlist(fd))

            fd.seek(self['header']['playlist_start_address'], os.SEEK_SET)
            self['playlist'] = as_normal_value(pyparsebluray.load_playlist(fd))



"""
Connect the user interface to the engine room
"""

from .base import JobBase  # isort:skip
from . import bdinfo, custom, dialog, imghost, mediainfo, playlists, poster, scene, screenshots, set, submit, torrent, webdb



"""
Find, download and re-upload poster for movie, series or season
"""

import collections
import hashlib
import os
import re
import urllib.parse

import async_lru

from .. import errors, uis, utils
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PosterJob(JobBase):
    """
    Get poster and optionally resize and reupload it
    """

    name = 'poster'
    label = 'Poster'

    # Caching is done by utilities.
    cache_id = None

    def initialize(self, *, getter, width=None, height=None, write_to=None, imghosts=()):
        """
        Set internal state

        :param getter: Coroutine function that returns a poster file or
                       poster URL (e.g. :meth:`.WebDbApiBase.poster_url`).
                       May raise :class:`~.RequestError`, which is passed
                       to :meth:`~.JobBase.error`.
        :param width: Resize poster to this many pixels wide (aspect ratio
                      is always maintained)
        :param height: Resize poster to this many pixels high (aspect
                       ratio always is maintained)
        :param imghosts: Upload poster to the first image hosting service in
                         this sequence of :class:`~.ImageHostBase` subclasses
                         and try the next one if this fails.
                         :class:`~.RequestError` from uploading is passed to
                         :meth:`~.JobBase.warn`. If all uploads fail,
                         :meth:`error` is called.
        :param write_to: Write poster to this file path (may be `None` or
                         empty string)
        """
        self._getter = getter
        self._width = width
        self._height = height
        self._write_to = write_to
        self._imghosts = imghosts

        self.signal.add('obtaining')
        self.signal.add('obtained')
        self.signal.add('downloading')
        self.signal.add('downloaded')
        self.signal.add('resizing')
        self.signal.add('resized')
        self.signal.add('uploading')
        self.signal.add('uploaded')

    _url_regex = re.compile(r'^https?://.+', flags=re.IGNORECASE)

    class _ProcessingError(errors.UpsiesError):
        pass

    async def run(self):
        try:
            params = await self._obtain()
            poster = params['poster']
            width = params.get('width', self._width)
            height = params.get('height', self._height)
            write_to = params.get('write_to', self._write_to)
            imghosts = params.get('imghosts', self._imghosts)

            poster = await self._resize(poster, width, height)
            await self._write(poster, write_to)
            await self._upload(poster, imghosts)

            if not write_to and not imghosts:
                if width or height:
                    await self._write(poster, self._get_poster_filename(poster))
                else:
                    self.add_output(poster)

        except self._ProcessingError as e:
            self.error(e)

    async def _obtain(self):
        _log.debug('Obtaining poster: %r', self._getter)
        self.signal.emit('obtaining')
        try:
            params_or_poster = await self._getter()
        except errors.RequestError as e:
            self.warn(f'Failed to get poster: {e}')
            params_or_poster = None

        _log.debug('Obtained poster: %r', params_or_poster)
        if not params_or_poster:
            params = await self._obtain_via_prompt()

        elif isinstance(params_or_poster, collections.abc.Mapping):
            params = params_or_poster

        else:
            params = {'poster': params_or_poster}

        self.signal.emit('obtained', params['poster'])
        return params

    async def _obtain_via_prompt(self):
        self.info = 'Please enter a poster file or URL.'
        try:
            poster = ''
            while True:
                poster = os.path.expanduser(await self.add_prompt(
                    uis.prompts.TextPrompt(text=poster)
                ))
                if not poster:
                    self.warn('Poster file or URL is required.')

                elif self._url_regex.search(poster):
                    # Download poster just to get an error if it fails. Later
                    # downloads should grab it from cache without pestering the
                    # server.
                    try:
                        await utils.http.get(poster, cache=True)
                    except errors.RequestError as e:
                        self.warn(f'Failed to download poster: {e}')
                    else:
                        return {'poster': poster}

                elif not os.path.exists(poster):
                    self.warn(f'Poster file does not exist: {poster}')

                elif not os.path.isfile(poster):
                    self.warn(f'Poster is not a file: {poster}')

                else:
                    return {'poster': poster}
        finally:
            self.clear_warnings()

    async def _resize(self, poster, width, height):
        if width or height:
            _log.debug('Resizing poster to %s x %s: %r', width, height, poster)

            # Download the poster so we can resize it.
            filepath = await self._get_poster_filepath(poster)
            filename_resized = '.'.join((
                utils.fs.basename(utils.fs.strip_extension(filepath)),
                f'{width}x{height}',
                utils.fs.file_extension(filepath),
            ))

            self.signal.emit('resizing', filepath)
            try:
                filepath_resized = utils.image.resize(
                    filepath,
                    target_directory=self.cache_directory,
                    target_filename=filename_resized,
                    width=width,
                    height=height,
                )
            except errors.ImageResizeError as e:
                raise self._ProcessingError(f'Failed to resize poster: {e}') from e
            else:
                self.signal.emit('resized', filepath_resized)
                return filepath_resized
        else:
            # Return original poster file or URL.
            return poster

    async def _write(self, poster, filepath):
        if filepath:
            # Write poster file or URL to user-provided path.
            _log.debug('Writing poster: %r', (poster, filepath))
            data = await self._read_file_or_url(poster)
            filepath = await self._write_file(data, filepath)
            self.add_output(filepath)

    async def _upload(self, poster, imghosts):
        if imghosts:
            # If poster is a URL, we must download it first.
            filepath = await self._get_poster_filepath(poster)

            # Upload `filepath` to any `imghost` and return the URL of the first
            # successful upload.
            for imghost in imghosts:
                _log.debug('Uploading poster: %r', (filepath, imghost.name))
                self.signal.emit('uploading', imghost)
                try:
                    url = await imghost.upload(filepath, thumb_width=0)
                except errors.RequestError as e:
                    self.warn(f'Failed to upload poster: {e}')
                else:
                    self.signal.emit('uploaded', url)
                    self.add_output(url)
                    return

            # If all upload() calls failed, add an error to all the warnings.
            raise self._ProcessingError('All uploads failed')

    @async_lru.alru_cache
    async def _get_poster_filepath(self, poster):
        if self._url_regex.search(poster):
            # Download poster URL to temporary file and return its path.
            data = await self._read_file_or_url(poster)
            filepath = os.path.join(
                self.cache_directory,
                self._get_poster_filename(poster),
            )
            return await self._write_file(data, filepath)

        else:
            # Return original poster file path.
            return poster

    def _get_poster_filename(self, poster):
        if self._url_regex.search(poster):
            # Turn poster URL into unique file name.
            url = urllib.parse.urlparse(poster)
            unique_id = hashlib.md5(
                '.'.join((url.path, url.query))
                .encode('utf8')
            ).hexdigest()
            filename = 'poster:' + '.'.join((
                url.hostname,
                unique_id,
            ))
            extension = utils.fs.file_extension(url.path)
            if extension:
                filename += f'.{extension}'
            return filename
        else:
            # Get poster file name from poster file path.
            return utils.fs.basename(poster)

    async def _read_file_or_url(self, poster):
        if self._url_regex.search(poster):
            # Return downloaded data from URL.
            _log.debug('Downloading poster: %r', poster)
            try:
                self.signal.emit('downloading', poster)
                response = await utils.http.get(poster, cache=True)
            except errors.RequestError as e:
                raise self._ProcessingError(f'Failed to download poster: {e}') from e
            else:
                self.signal.emit('downloaded', poster)
                return response.bytes

        else:
            # Return poster file content.
            _log.debug('Reading poster: %r', poster)
            try:
                with open(poster, 'rb') as f:
                    return f.read()
            except OSError as e:
                msg = e.strerror if e.strerror else str(e)
                raise self._ProcessingError(f'Failed to read poster: {msg}') from e

    async def _write_file(self, data, filepath):
        # Write `data` to sanitized `filepath` and return sanitized `filepath`.
        filepath = utils.fs.sanitize_path(filepath)
        try:
            with open(filepath, 'wb') as f:
                f.write(data)
        except OSError as e:
            msg = e.strerror if e.strerror else str(e)
            raise self._ProcessingError(f'Failed to write {filepath}: {msg}') from e
        else:
            return filepath



"""
Subtitles information
"""

import functools
import re

from ... import errors, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)

langcodes = utils.LazyModule(module='langcodes', namespace=globals())


# Map various strings to canonical sutitle format names
_subtitle_formats = {
    # Matroska Codec ID
    # Reference: https://www.matroska.org/technical/codec_specs.html
    'S_HDMV/PGS': 'PGS',     # Blu-ray
    'S_VOBSUB': 'VobSub',    # DVD
    'S_TEXT/UTF8': 'SRT',    # SubRip
    'S_TEXT/SSA': 'SSA',     # SubStation Alpha
    'S_TEXT/ASS': 'ASS',     # Advanced SubStation Alpha
    'S_TEXT/WEBVTT': 'VTT',  # Web Video Text Tracks

    # File extension
    'srt': 'SRT',     # SubRip
    'ssa': 'SSA',     # SubStation Alpha
    'ass': 'ASS',     # Advanced SubStation Alpha
    'vtt': 'VTT',     # Web Video Text Tracks
    'idx': 'VobSub',  # DVD
    'sub': 'VobSub',  # DVD
}


class Subtitle:
    """
    Dataclass that stores information about a subtitle track

    :param language: BCP47 language code with optional country code (e.g. "en",
        "en-UK", "es-419", "zh-HANS")

        References:
            https://www.matroska.org/technical/notes.html [Language Codes]
            https://en.wikipedia.org/wiki/IETF_language_tag

        If `language` is not a valid language code, the `language` and
        `region` attributes are empty strings.

    :param bool forced: Whether the subtitle track should always be played
    :param bool format: Subtitle format (e.g. "SRT", "VobSub", "PGS")

    :raise ValueError: if `language` is invalid
    """

    def __init__(self, *, language, forced, format):
        # Parse and normalize `language`
        try:
            lc = langcodes.Language.get(language)
        except ValueError:
            self._language = ''
            self._region = ''
        else:
            self._language = lc.language
            self._region = lc.territory or ''

        self._forced = bool(forced)
        self._format = str(format or '')

    @property
    def language(self):
        """
        Language part of the `language` initialization argument, e.g. "en",
        "es", "zh", etc.
        """
        return self._language

    @property
    def region(self):
        """
        Region part of the `language` initialization argument, e.g. "UK",
        "419", "HANS", etc.
        """
        return self._region

    @property
    def forced(self):
        """Whether the subtitle track should always be played"""
        return self._forced

    @property
    def format(self):
        """Subtitle format (e.g. "SRT", "VobSub", "PGS")"""
        return self._format

    def __repr__(self):
        if self.region:
            language = f'{self.language}-{self.region}'
        else:
            language = self.language
        return f'{type(self).__name__}(language={language!r}, forced={self.forced!r}, format={self.format!r})'

    def __hash__(self):
        return hash((self.language, self.region, self.forced, self.format))

    def __eq__(self, other):
        return hash(self) == hash(other)


def get_subtitles(content_path):
    """
    Return sequence of :class:`Subtitle` objects from all subtitle sources

    This function combines all ``get_subtitles_*`` functions.

    If `content_path` is a multi-disc release (i.e. a directory that contains multiple directories
    that each contain a VIDEO_TS or BDMV subdirectory), it returns the combined subtitles of all
    discs.

    :param content_path: Path to release file or directory
    """
    subtitles = []

    if utils.disc.is_bluray(content_path, multidisc=True):
        # BDMV
        for disc_path in utils.disc.bluray.get_disc_paths(content_path):
            subtitles.extend(get_subtitles_from_bluray(disc_path))

    elif utils.disc.is_dvd(content_path, multidisc=True):
        # VIDEO_TS
        for disc_path in utils.disc.dvd.get_disc_paths(content_path):
            subtitles.extend(get_subtitles_from_dvd(disc_path))

    else:
        # Muxed subtitles in video file(s).
        subtitles.extend(get_subtitles_from_mediainfo(content_path))
        # Separate srt/ass/ssa files in `content_path` directory.
        subtitles.extend(get_subtitles_from_text_files(content_path))
        # Separate idx/sub files in `content_path` directory.
        subtitles.extend(get_subtitles_from_idx_files(content_path))

    return tuple(subtitles)


def get_subtitles_from_mediainfo(content_path):
    """
    Return sequence of :class:`Subtitle` objects from ``mediainfo`` output

    Only the first video is used to find subtitles (see :func:`find_videos`).

    :param content_path: Path to release file or directory
    """
    subtitles = []
    try:
        subtitle_tracks = utils.mediainfo.lookup(content_path, ('Text',))
    except errors.ContentError as e:
        _log.debug('Failed to lookup subtitle tracks: %r', e)
        pass
    else:
        subtitles = []
        for track in subtitle_tracks:
            subtitle = Subtitle(
                language=track.get('Language', ''),
                forced=(track.get('Forced') == 'Yes'),
                format=_subtitle_formats.get(track.get('CodecID')),
            )
            _log.debug('Found subtitle language in mediainfo: %r', subtitle)
            subtitles.append(subtitle)
    _log.debug('Subtitles from mediainfo: %r', subtitles)
    return tuple(subtitles)


_srt_filename_regex = re.compile(r'(?:\.([a-z]{2,3})|)\.(\w+)$')

@functools.cache
def get_subtitles_from_text_files(content_path):
    """
    Return sequence of :class:`Subtitle` objects from subtitle file names

    For a video file named "foo.mkv", the subtitle language can be in the file
    name, e.g. "foo.en.srt" or "foo.eng.ass".

    If `content_path` is a directory, it is searched recursively for subtitle
    files.

    If `content_path` is not a directory, the returned sequence is empty.

    :param content_path: Path to release file or directory
    """
    subtitles = []
    for filepath in utils.fs.file_list(content_path, extensions=('srt', 'ssa', 'ass', 'vtt')):
        filename = utils.fs.basename(filepath)
        match = _srt_filename_regex.search(filename)

        if match:
            language = match.group(1) or ''
            extension = match.group(2).lower()
            format = _subtitle_formats.get(extension, '')
        else:
            language = ''
            format = ''

        subtitle = Subtitle(
            language=language,
            forced=False,
            format=format,
        )
        _log.debug('Found text subtitle: %s: %r', filepath, subtitle)
        subtitles.append(subtitle)

    _log.debug('Subtitles from text files: %r', subtitles)
    return subtitles


@functools.cache
def get_subtitles_from_idx_files(content_path):
    """
    Return sequence of :class:`Subtitle` objects from ``*.idx`` files

    For .idx/.sub pairs, the .idx file can contain the language.

    If `content_path` is a directory, it is searched recursively for ``*.idx``
    files.

    If `content_path` is not a directory, the returned sequence is empty.

    :param content_path: Path to release file or directory
    """
    subtitles = []
    for idx_filepath in utils.fs.file_list(content_path, extensions=('idx',)):
        try:
            language = _get_language_code_from_idx_file(idx_filepath)
        except OSError as e:
            _log.debug('Ignoring exception from reading %r: %r', idx_filepath, e)
        else:
            subtitle = Subtitle(
                language=language,
                forced=False,
                format=_subtitle_formats['idx'],
            )
            _log.debug('Found IDX subtitle: %s: %r', idx_filepath, subtitle)
            subtitles.append(subtitle)
    _log.debug('Subtitles from idx files: %r', subtitles)
    return subtitles


_idx_language_regex = re.compile(r'^id:\s*([a-z]{2,3})\b')

def _get_language_code_from_idx_file(idx_filepath):
    # Example: "id: de, index: 0"
    # Expect 2-letter and 3-letter codes.
    with open(idx_filepath, 'r') as f:
        for line in f:
            match = _idx_language_regex.search(line)
            if match:
                language = match.group(1)
                _log.debug('Found subtitle language in %s: %r', idx_filepath, language)
                return language
    return ''


@functools.cache
def get_subtitles_from_dvd(content_path):
    """
    Return sequence of :class:`Subtitle` objects from ``VIDEO_TS`` subdirectory

    If `content_path` is not a directory or does not contain a ``VIDEO_TS`` subdirectory, the
    returned sequence is empty.

    :param content_path: Path directory that contains ``VIDEO_TS`` subdirectory
    """
    subtitles = []
    main_playlists = tuple(
        playlist
        for playlist in utils.disc.dvd.get_playlists(content_path)
        if playlist.is_main
    )
    for playlist in main_playlists:
        for sub in get_subtitles_from_mediainfo(playlist.filepath):
            sub._format = _subtitle_formats['S_VOBSUB']
            subtitles.append(sub)
    _log.debug('Subtitles from DVD tree: %r', subtitles)
    return subtitles


@functools.cache
def get_subtitles_from_bluray(content_path):
    """
    Return sequence of :class:`Subtitle` objects from ``BDMV`` subdirectory

    If `content_path` is not a directory or does not contain a ``BDMV`` subdirectory, the
    returned sequence is empty.

    :param content_path: Path directory that contains ``BDMV`` subdirectory
    """
    subtitles = []
    main_playlists = tuple(
        playlist
        for playlist in utils.disc.bluray.get_playlists(content_path)
        if playlist.is_main
    )
    for playlist in main_playlists:
        for sub in get_subtitles_from_mediainfo(playlist.filepath):
            sub._format = _subtitle_formats['S_HDMV/PGS']
            subtitles.append(sub)
    _log.debug('Subtitles from Blu-ray tree: %r', subtitles)
    return subtitles



import sys

import logging  # isort:skip
_log = logging.getLogger(__name__)


def print_stdout(*msgs):
    """Simplified version of regular :func:`print` that prints to :attr:`sys.stdout`"""
    sys.stdout.write(' '.join(str(msg) for msg in msgs) + '\n')


def print_stderr(*msgs):
    """Simplified version of regular :func:`print` that prints to :attr:`sys.stderr`"""
    sys.stderr.write(' '.join(str(msg) for msg in msgs) + '\n')


def is_tty():
    """Whether we live in a terminal and can interact with the user"""
    # If we have input and output, we have user interaction. It's ok if we don't have sys.stdout
    # (e.g. because it is redirected to a file) because prompt-toolkit writes to sys.stderr in that
    # case.
    return bool(
        # stdin does exist and is not redirected
        (sys.stdin and sys.stdin.isatty())
        and (
            # stdout exists and is not redirected.
            (sys.stdout and sys.stdout.isatty())
            # stderr exists and is not redirected.
            or (sys.stderr and sys.stderr.isatty())
        )
    )



"""
Abstract base class for tracker jobs
"""

import abc
import builtins
import collections
import functools
import pathlib
import re

from ... import __homepage__, __project_name__, errors, jobs, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


class TrackerJobsBase(abc.ABC):
    """
    Base class for tracker-specific :class:`jobs <upsies.jobs.base.JobBase>`

    This base class defines general-purpose jobs that can be used by subclasses
    by returning them in their :attr:`jobs_before_upload` or
    :attr:`jobs_after_upload` attributes. It also provides all objects that are
    needed by any one of those jobs.

    Job instances are provided as :func:`functools.cached_property`, i.e. jobs
    are created only once per session.

    Subclasses that need to run background tasks should pass them to
    :meth:`.JobBase.add_task` or to :meth:`.TrackerBase.attach_task`.
    :meth:`.TrackerBase.attach_task` should only be used if there is no
    appropriate job for the task.

    For a description of the arguments see the corresponding properties.
    """

    def __init__(self, *, content_path, tracker,
                 reuse_torrent_path=None, exclude_files=(),
                 btclient=None, torrent_destination=None,
                 screenshots_optimization=None, image_hosts=None,
                 show_poster=False,
                 options=None, common_job_args=None):
        self._content_path = content_path
        self._tracker = tracker
        self._reuse_torrent_path = reuse_torrent_path
        self._exclude_files = exclude_files
        self._btclient = btclient
        self._torrent_destination = torrent_destination
        self._screenshots_optimization = screenshots_optimization
        self._image_hosts = image_hosts
        self._show_poster = show_poster
        self._options = options or {}
        self._common_job_args = common_job_args or {}
        self._signal = utils.signal.Signal(
            id=f'{self.tracker.name}-tracker-jobs',
            signals=('warning', 'error', 'exception'),
        )

    @property
    def content_path(self):
        """
        Content path to generate metadata for

        This is the same object that was passed as initialization argument.
        """
        return self._content_path

    @property
    def tracker(self):
        """
        :class:`~.trackers.base.TrackerBase` subclass

        This is the same object that was passed as initialization argument.
        """
        return self._tracker

    @property
    def reuse_torrent_path(self):
        """
        Path to an existing torrent file that matches :attr:`content_path`

        See :func:`.torrent.create`.
        """
        return self._reuse_torrent_path

    @property
    def torrent_destination(self):
        """
        Where to copy the generated torrent file to or `None`

        This is the same object that was passed as initialization argument.
        """
        return self._torrent_destination

    @property
    def exclude_files(self):
        """
        Sequence of glob and regular expression patterns to exclude from the
        generated torrent

        See the ``exclude_files`` argument of
        :meth:`.CreateTorrentJob.initialize`.

        This should also be used by :attr:`screenshots_job` to avoid making
        screenshots from files that aren't in the torrent.
        """
        return self._exclude_files

    @property
    def options(self):
        """
        Configuration options provided by the user

        This is the same object that was passed as initialization argument.
        """
        return self._options

    @property
    def image_hosts(self):
        """
        Sequence of :class:`~.base.ImageHostBase` instances or `None`

        This is the same object that was passed as initialization argument.
        """
        return self._image_hosts

    @property
    def btclient(self):
        """
        :class:`~.btclient.BtClient` instance or `None`

        This is the same object that was passed as initialization argument.
        """
        return self._btclient

    @functools.cached_property
    def is_bdmv_release(self):
        """Whether :attr:`content_path` is a Blu-ray release (see :func:`~.is_bluray`)"""
        return utils.disc.is_bluray(self.content_path, multidisc=True)

    @functools.cached_property
    def is_video_ts_release(self):
        """Whether :attr:`content_path` is a DVD release (see :func:`~.is_dvd`)"""
        return utils.disc.is_dvd(self.content_path, multidisc=True)

    @functools.cached_property
    def is_disc_release(self):
        """Whether :attr:`content_path` is a Blu-ray or DVD release (see :func:`~.is_disc`)"""
        return utils.disc.is_disc(self.content_path, multidisc=True)

    def common_job_args(self, **overload):
        """
        Keyword arguments that are passed to all jobs or empty `dict`

        :param overload: Keyword arguments add or replace values from the
            initialization argument
        """
        # Combine global defaults with custom values.
        args = {
            **self._common_job_args,
            **overload,
        }

        # Individual jobs may only set `ignore_cache=False` if
        # `ignore_cache=True` wasn't set globally, e.g. with --ignore-cache.
        if self._common_job_args.get('ignore_cache') or overload.get('ignore_cache'):
            args['ignore_cache'] = True

        return args

    @property
    @abc.abstractmethod
    def jobs_before_upload(self):
        """
        Sequence of jobs that need to finish before :meth:`~.TrackerBase.upload` can
        be called
        """

    @functools.cached_property
    def jobs_after_upload(self):
        """
        Sequence of jobs that are started after :meth:`~.TrackerBase.upload`
        finished

        .. note:: Jobs returned by this class should have
                  :attr:`~.JobBase.autostart` set to `False` or they will be
                  started before submission is attempted.

        By default, this returns :attr:`add_torrent_job` and
        :attr:`copy_torrent_job`.
        """
        return (
            self.add_torrent_job,
            self.copy_torrent_job,
        )

    @property
    def isolated_jobs(self):
        """
        Sequence of job names (e.g. ``"imdb_job"``) that were singled out by the
        user (e.g. with a CLI argument) to create only a subset of the usual
        metadata

        If this sequence is empty, all jobs in :attr:`jobs_before_upload` and
        :attr:`jobs_after_upload` are enabled.
        """
        return ()

    def get_job_and_dependencies(self, *jobs):
        """
        Combine all `jobs` and their dependencies recursively into flat sequence

        :param jobs: :class:`~.JobBase` instances

        Dependencies are gathered from each job's :attr:`~.JobBase.prejobs` and
        :meth:`~.JobBase.presignal`\\ s.

        .. warning:: This is not a foolproof way to find all dependencies.

                     Jobs may :meth:`~.JobBase.receive_all` signals from :attr:`~.JobBase.siblings`,
                     but only optionally (e.g. :class:`~.AddTorrentJob`), so they don't specify the
                     required siblings as dependencies.

        :return: Sequence of :class:`~.JobBase` instances
        """
        all_jobs = set()

        def add_job(job):
            if job not in all_jobs:
                all_jobs.add(job)

        def add_prejobs(job):
            for prejob in job.prejobs:
                add_job(prejob)
                add_prejobs(prejob)
                add_jobs_from_presignals(prejob)

        def add_jobs_from_presignals(job):
            for presignal_job in job.presignals:
                add_job(presignal_job)
                add_prejobs(presignal_job)
                add_jobs_from_presignals(presignal_job)

        for j in jobs:
            add_job(j)
            add_prejobs(j)
            add_jobs_from_presignals(j)

        return tuple(all_jobs)

    @property
    def submission_ok(self):
        """
        Whether the created metadata should be submitted

        The base class implementation returns `False` if there are any
        :attr:`isolated_jobs`. Otherwise, it returns `True` only if all
        :attr:`jobs_before_upload` have an :attr:`~.base.JobBase.exit_code` of
        ``0`` or a falsy :attr:`~.base.JobBase.is_enabled` value.

        Subclasses should always call the parent class implementation to ensure
        all metadata was created successfully.
        """
        if self.isolated_jobs:
            # If some jobs are disabled, required metadata is missing and we
            # can't submit
            return False
        else:
            enabled_jobs_before_upload = tuple(
                job for job in self.jobs_before_upload
                if job and job.is_enabled
            )
            enabled_jobs_succeeded = all(
                job.exit_code == 0
                for job in enabled_jobs_before_upload
            )
            return bool(
                enabled_jobs_before_upload
                and enabled_jobs_succeeded
            )

    @property
    def signal(self):
        """
        :class:`~.signal.Signal` instance with the signals ``warning``, ``error``
        and ``exception``
        """
        return self._signal

    def warn(self, warning):
        """
        Emit ``warning`` signal (see :attr:`signal`)

        Emit a warning for any non-critical issue that the user can choose to
        ignore or fix.
        """
        self.signal.emit('warning', warning)

    def error(self, error):
        """
        Emit ``error`` signal (see :attr:`signal`)

        Emit an error for any critical but expected issue that can't be
        recovered from (e.g. I/O error).
        """
        self.signal.emit('error', error)

    def exception(self, exception):
        """
        Emit ``exception`` signal (see :attr:`signal`)

        Emit an exception for any critical and unexpected issue that should be
        reported as a bug.
        """
        self.signal.emit('exception', exception)

    @functools.cached_property
    def imdb(self):
        """:class:`~.webdbs.imdb.ImdbApi` instance"""
        return utils.webdbs.webdb('imdb')

    @functools.cached_property
    def tmdb(self):
        """:class:`~.webdbs.tmdb.TmdbApi` instance"""
        return utils.webdbs.webdb('tmdb')

    @functools.cached_property
    def tvmaze(self):
        """:class:`~.webdbs.tvmaze.TvmazeApi` instance"""
        return utils.webdbs.webdb('tvmaze')

    def get_job_name(self, name):
        """
        Return job name that is unique for this tracker

        It's important for tracker jobs to have unique names to avoid re-using
        cached output from another tracker's job with the same name.

        Standard jobs have names so that cached output will be re-used by other
        trackers if possible. This function is mainly for unique and custom jobs
        that are only used for one tracker but might share the same name with
        other trackers.
        """
        suffix = f'.{self.tracker.name}'
        if name.endswith(suffix):
            return name
        else:
            return f'{name}{suffix}'

    @functools.cached_property
    def create_torrent_job(self):
        """:class:`~.jobs.torrent.CreateTorrentJob` instance"""
        return jobs.torrent.CreateTorrentJob(
            content_path=self.content_path,
            reuse_torrent_path=self.reuse_torrent_path,
            tracker=self.tracker,
            exclude_files=self.exclude_files,
            precondition=self.make_precondition('create_torrent_job'),
            **self.common_job_args(),
        )

    @functools.cached_property
    def add_torrent_job(self):
        """:class:`~.jobs.torrent.AddTorrentJob` instance"""
        if self._btclient and self.create_torrent_job:
            return jobs.torrent.AddTorrentJob(
                autostart=False,
                btclient=self._btclient,
                precondition=self.make_precondition('add_torrent_job'),
                **self.common_job_args(),
            )

    @functools.cached_property
    def copy_torrent_job(self):
        """:class:`~.jobs.torrent.CopyTorrentJob` instance"""
        if self.torrent_destination and self.create_torrent_job:
            return jobs.torrent.CopyTorrentJob(
                autostart=False,
                destination=self.torrent_destination,
                precondition=self.make_precondition('copy_torrent_job'),
                **self.common_job_args(),
            )

    @property
    def torrent_filepath(self):
        """Local path to the torrent file created by :attr:`create_torrent_job`"""
        return self.get_job_output(self.create_torrent_job, slice=0)

    @functools.cached_property
    def subtitles(self):
        """Sequence of :mod:`~.mediainfo.text.Subtitle` objects for :attr:`content_path`"""
        return utils.mediainfo.text.get_subtitles(self.content_path)

    @functools.cached_property
    def release_name(self):
        """
        :class:`~.release.ReleaseName` instance with
        :attr:`release_name_translation` applied
        """
        return utils.release.ReleaseName(
            path=self.content_path,
            translate=self.release_name_translation,
            separator=self.release_name_separator,
            english_title_before_original=self.release_name_english_title_before_original,
        )

    release_name_separator = None
    """See :attr:`.ReleaseName.separator`"""

    release_name_english_title_before_original = False
    """See :attr:`~.utils.release.ReleaseName.english_title_before_original`"""

    release_name_translation = {}
    """See ``translate`` argument of :class:`~.utils.release.ReleaseName`"""

    @functools.cached_property
    def release_name_job(self):
        """
        :class:`~.jobs.dialog.TextFieldJob` instance with text set to
        :attr:`release_name`

        The text is automatically updated when :attr:`imdb_job` sends an ID.
        """
        # NOTE: This job should not use the same cache as the `release-name`
        #       subcommand because a tracker's release_name_job can make
        #       arbitrary customizations.
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('release-name'),
            label='Release Name',
            callbacks={
                'output': self.release_name.set_release_info,
            },
            precondition=self.make_precondition('release_name_job'),
            validator=self.validate_release_name,
            **self.common_job_args(),
        )

    def validate_release_name(self, text):
        if not text.strip():
            raise ValueError('Release name must not be empty.')

        match = re.search(rf'(?:{utils.release.DELIM}|$)(UNKNOWN_([A-Z_]+))(?:{utils.release.DELIM}|$)', text)
        if match:
            placeholder = match.group(1)
            attribute = match.group(2).replace('_', ' ')
            raise ValueError(f'Replace "{placeholder}" with the proper {attribute.lower()}.')

    async def update_release_name_from(self, webdb, webdb_id):
        """
        Update :attr:`release_name_job` with web DB information

        :param webdb: :class:`~.webdbs.base.WebDbApiBase` instance
        :param webdb_id: ID for `webdb`

        This is a convenience wrapper around :meth:`.ReleaseName.fetch_info` and
        :meth:`.TextFieldJob.fetch_text`.
        """
        if not self.release_name_job.is_finished:
            await self.release_name_job.fetch_text(
                coro=self.release_name.fetch_info(webdb=webdb, webdb_id=webdb_id),
                nonfatal_exceptions=(errors.RequestError,),
                default=str(self.release_name),
            )
            _log.debug('Updated release name: %s', self.release_name)
        else:
            _log.debug('Not updating release name because %r is finished', self.release_name_job)

    @functools.cached_property
    def imdb_job(self):
        """:class:`~.jobs.webdb.WebDbSearchJob` instance"""
        return jobs.webdb.WebDbSearchJob(
            query=self.content_path,
            db=self.imdb,
            autodetect=self.autodetect_imdb_id,
            show_poster=self._show_poster,
            callbacks={
                'output': self._handle_imdb_id,
            },
            precondition=self.make_precondition('imdb_job'),
            **self.common_job_args(),
        )

    async def autodetect_imdb_id(self):
        # Get ID from CLI option or other user-provided source.
        if self.options.get('imdb'):
            _log.debug('Found IMDb ID in CLI: %r', self.options)
            return self.options['imdb']

        # Get ID from video container tag.
        imdb_id = utils.mediainfo.lookup(
            path=self.content_path,
            keys=('General', 0, 'extra', 'IMDB'),
            default=None,
        )
        if imdb_id:
            _log.debug('Found IMDb ID in mediainfo: %r', imdb_id)
            return imdb_id

    def _handle_imdb_id(self, imdb_id):
        # Update other webdb queries with IMDb info
        self.tracker.attach_task(self._propagate_webdb_info(self.imdb, imdb_id))

    @property
    def imdb_id(self):
        """IMDb ID if :attr:`imdb_job` is finished or `None`"""
        return self.imdb_job.selected.get('id', None)

    @functools.cached_property
    def tmdb_job(self):
        """:class:`~.jobs.webdb.WebDbSearchJob` instance"""
        return jobs.webdb.WebDbSearchJob(
            query=self.content_path,
            db=self.tmdb,
            autodetect=self.autodetect_tmdb_id,
            show_poster=self._show_poster,
            callbacks={
                'output': self._handle_tmdb_id,
            },
            precondition=self.make_precondition('tmdb_job'),
            **self.common_job_args(),
        )

    async def autodetect_tmdb_id(self):
        # Get ID from CLI option or other user-provided source.
        if self.options.get('tmdb'):
            _log.debug('Found TMDb ID in CLI: %r', self.options)
            return self.options['tmdb']

        # Get ID from video container tag.
        tmdb_id = utils.mediainfo.lookup(
            path=self.content_path,
            keys=('General', 0, 'extra', 'TMDB'),
            default=None,
        )
        if tmdb_id:
            _log.debug('Found TMDb ID in mediainfo: %r', tmdb_id)
            return tmdb_id

    def _handle_tmdb_id(self, tmdb_id):
        # Do NOT update other webdb queries with TMDb info. TMDb year seems to
        # change based on geolocation and is therefore garbage.
        # self.tracker.attach_task(self._propagate_webdb_info(self.tmdb, tmdb_id))
        pass

    @property
    def tmdb_id(self):
        """TMDb ID if :attr:`tmdb_job` is finished or `None`"""
        return self.tmdb_job.selected.get('id', None)

    @functools.cached_property
    def tvmaze_job(self):
        """:class:`~.jobs.webdb.WebDbSearchJob` instance"""
        return jobs.webdb.WebDbSearchJob(
            query=self.content_path,
            db=self.tvmaze,
            autodetect=self.autodetect_tvmaze_id,
            show_poster=self._show_poster,
            callbacks={
                'output': self._handle_tvmaze_id,
            },
            precondition=self.make_precondition('tvmaze_job'),
            **self.common_job_args(),
        )

    async def autodetect_tvmaze_id(self):
        # Get ID from CLI option or other user-provided source.
        if self.options.get('tvmaze'):
            _log.debug('Found TVmaze ID in CLI: %r', self.options)
            return self.options['tvmaze']

        # Get ID from video container tag.
        tvmaze_id = utils.mediainfo.lookup(
            path=self.content_path,
            keys=('General', 0, 'extra', 'TVMAZE'),
            default=None,
        )
        if tvmaze_id:
            _log.debug('Found TVmaze ID in mediainfo: %r', tvmaze_id)
            return tvmaze_id

    def _handle_tvmaze_id(self, tvmaze_id):
        # Update other webdb queries with TVmaze info
        self.tracker.attach_task(self._propagate_webdb_info(self.tvmaze, tvmaze_id))

    @property
    def tvmaze_id(self):
        """TVmaze ID if :attr:`tvmaze_job` is finished or `None`"""
        return self.tvmaze_job.selected.get('id', None)

    async def _propagate_webdb_info(self, webdb, webdb_id):
        target_webdb_jobs = [
            j for j in (getattr(self, f'{name}_job') for name in utils.webdbs.webdb_names())
            if (
                webdb.name not in j.name
                and j.is_enabled
                and not j.is_finished
            )
        ]

        if target_webdb_jobs:
            title_english = await webdb.title_english(webdb_id)
            title_original = await webdb.title_original(webdb_id)
            query = utils.webdbs.Query(
                type=await webdb.type(webdb_id),
                title=title_english or title_original,
                year=await webdb.year(webdb_id),
            )

            _log.debug('Propagating %s info to other webdbs: %r: %s', webdb.name, [j.name for j in target_webdb_jobs], query)
            for job in target_webdb_jobs:
                job.query.update(query)

        await self.update_release_name_from(webdb, webdb_id)

    @functools.cached_property
    def screenshots_job(self):
        """
        :class:`~.jobs.screenshots.ScreenshotsJob` instance

        The number of screenshots to make is taken the :attr:`screenshots_count`
        attribute.
        """
        return jobs.screenshots.ScreenshotsJob(
            content_path=self.content_path,
            precreated=self.screenshots_precreated,
            exclude_files=self.exclude_files,
            count=self.screenshots_count,
            from_all_videos=self.document_all_videos,
            optimize=self._screenshots_optimization,
            precondition=self.make_precondition('screenshots_job'),
            **self.common_job_args(),
        )

    @property
    def screenshots_precreated(self):
        """
        Sequence of user-provided screenshot file paths

        The default implementation uses :attr:`options`\\ ``["screenshots"]``.
        It may be an arbitrarily nested list, which is flattened.
        """
        return utils.flatten_nested_lists(
            self.options.get('screenshots', ())
        )

    @property
    def screenshots_count(self):
        """
        How many screenshots to make

        The default implementation uses :attr:`options`\\ ``["screenshots_count"]``
        with `None` as the default value, which creates a default number of
        screenshots.
        """
        return self.options.get('screenshots_count')

    @property
    def document_all_videos(self):
        """
        Whether to document all videos in the release or just the first/main video

        Documenting means creating screenshots and a Mediainfo or BDInfo report.

        Extras, and samples should never be documented even if this is enabled. (This is tricky and
        may fail sometimes.)

        For movies and episodes, this has no effect. For seasons, this documents each episode. For
        Blu-rays and DVDs, document each playlist that is selected by the user.

        The default is `False`.
        """
        return False

    image_host_config = {}
    """
    Dictionary that maps an image hosting service :attr:`~.ImageHostBase.name`
    to :attr:`~.ImageHostBase.default_config` values

    ``common`` is a special image host whose values are always applied.

    Values from a specific image hosting service overload ``common`` values.

    Example:

    >>> image_host_config = {
    ...     # Always generate 300p thumbnails
    ...     "common": {"thumb_width": 300},
    ...     # Use API key for specific image hosting service
    ...     "myhost": {"apikey": "d34db33f"},
    ... }
    """

    @functools.cached_property
    def upload_screenshots_job(self):
        """:class:`~.jobs.imghost.ImageHostJob` instance"""
        if self.image_hosts and self.screenshots_job:
            return jobs.imghost.ImageHostJob(
                imghosts=self.image_hosts,
                precondition=self.make_precondition('upload_screenshots_job'),
                **self.common_job_args(),
            )

    poster_max_width = 300
    """Maximum poster image width"""

    poster_max_height = 600
    """Maximum poster image height"""

    @functools.cached_property
    def poster_job(self):
        """
        :class:`~.jobs.poster.PosterJob` instance

        See also :meth:`get_poster`, :meth:`get_poster_from_user`,
        and :meth:`get_poster_from_webdb`.
        """
        return jobs.poster.PosterJob(
            precondition=self.make_poster_job_precondition(),
            getter=self.get_poster,
            width=self.poster_max_width,
            height=self.poster_max_height,
            write_to=None,
            imghosts=self.image_hosts,
            **self.common_job_args(),
        )

    def make_poster_job_precondition(self):
        """
        :attr:`~.JobBase.precondition` for :attr:`poster_job`

        Subclasses may override this method to selectively provide a poster only
        if the server doesn't have one yet.
        """
        return self.make_precondition('poster_job')

    async def get_poster(self):
        """
        Return poster file or URL or `None`

        The default implementation tries to get the poster from the following
        methods and returns the first truthy return value:

            - :meth:`get_poster_from_user`
            - :meth:`get_poster_from_tracker`
            - :meth:`get_poster_from_webdb`

        Besides a file or URL, the return value may also be a dictionary with
        the key ``poster`` and the following optional keys:

            - ``width`` - Resize width in pixels (keep aspeect ratio)
            - ``height`` - Resize height in pixels (keep aspeect ratio)
            - ``write_to`` - Write resized poster to this file
            - ``imghosts`` - Sequence of :class:`~.ImageHostBase` instances to
              try to upload the poster to

        See :class:`~.PosterJob` for more information.
        """
        poster = await self.get_poster_from_user()
        if poster:
            return poster

        poster = await self.get_poster_from_tracker()
        if poster:
            return poster

        poster = await self.get_poster_from_webdb()
        if poster:
            return poster

    async def get_poster_from_user(self):
        """
        Get poster from user (e.g. CLI argument)

        The default implementation uses :attr:`options`\\ ``["poster"]``.
        """
        return self.options.get('poster', None)

    async def get_poster_from_tracker(self):
        """
        Get poster from tracker or any other custom source

        The default implementation always returns `None`.
        """
        return None

    async def get_poster_from_webdb(self):
        """Return poster URL from :attr:`poster_webdb` or `None`"""
        if self.poster_webdb_job:
            # We can't pass self.poster_webdb_job via `prejobs` to PosterJob because
            # self.poster_webdb_job is determined by checking which webdb job (self.imdb_job,
            # self.tvmaze_job, etc) is contained in self.jobs_before_upload. But
            # self.jobs_before_upload contains self.poster_job, resulting in infinite recursion.
            await self.poster_webdb_job.wait_finished()
            # Because imdb_job.no_id_ok may be True, we have to handle poster_webdb_job.output being
            # empty.
            webdb_id = self.get_job_output(self.poster_webdb_job, slice=0, default=None)
            if webdb_id:
                try:
                    poster = await self.poster_webdb.poster_url(
                        webdb_id,
                        season=self.release_name.only_season,
                    )
                except errors.RequestError as e:
                    _log.debug('Failed to get poster from %s: %r', self.poster_webdb, e)
                else:
                    if poster:
                        return poster

    @functools.cached_property
    def poster_webdb_job(self):
        """
        :class:`~.jobs.base.WebDbSearchJob` instance that is used by
        :meth:`get_poster_from_webdb` to get a poster image or `None` if no
        such instance is enabled and contained in :attr:`jobs_before_upload`
        """
        _webdb, job = self._poster_webdb_and_job
        return job

    @functools.cached_property
    def poster_webdb(self):
        """
        :class:`~.webdbs.base.WebDbApiBase` instance that is used by
        :meth:`get_poster_from_webdb` to get a poster image or `None` if no
        such instance is enabled and contained in :attr:`jobs_before_upload`
        """
        webdb, _job = self._poster_webdb_and_job
        return webdb

    @property
    def _poster_webdb_and_job(self):
        if (
                self.poster_job.is_enabled
                and self.poster_job in self.jobs_before_upload
        ):
            if (
                    self.tvmaze_job.is_enabled
                    and self.tvmaze_job in self.jobs_before_upload
                    and self.release_name.type in (utils.release.ReleaseType.season,
                                                   utils.release.ReleaseType.episode)
            ):
                return self.tvmaze, self.tvmaze_job

            elif (
                    self.imdb_job.is_enabled
                    and self.imdb_job in self.jobs_before_upload
            ):
                return self.imdb, self.imdb_job

            elif (
                    self.tmdb_job.is_enabled
                    and self.tmdb_job in self.jobs_before_upload
            ):
                return self.tmdb, self.tmdb_job

        return None, None

    @functools.cached_property
    def playlists_job(self):
        """:class:`~.jobs.playlists.PlaylistsJob` instance"""
        return jobs.playlists.PlaylistsJob(
            content_path=self.content_path,
            select_multiple=self.document_all_videos,
            precondition=self.make_precondition('playlists_job', precondition=self.playlists_job_precondition),
            **self.common_job_args(),
        )

    def playlists_job_precondition(self):
        return utils.disc.is_disc(self.content_path, multidisc=True)

    @functools.cached_property
    def mediainfo_job(self):
        """:class:`~.jobs.mediainfo.MediainfoJob` instance"""
        return jobs.mediainfo.MediainfoJob(
            content_path=self.content_path,
            from_all_videos=self.document_all_videos,
            exclude_files=self.exclude_files,
            precondition=self.make_precondition('mediainfo_job', precondition=self.mediainfo_job_precondition),
            **self.common_job_args(),
        )

    def mediainfo_job_precondition(self):
        return (
            # Non-BDMV releases always require mediainfo.
            not self.is_bdmv_release
            # BDMV releases require mediainfo based on the specific tracker.
            or self.mediainfo_required_for_bdmv
        )

    mediainfo_required_for_bdmv = False
    """Whether the tracker requires ``mediainfo`` output for BDMV Blu-ray releases"""

    @functools.cached_property
    def bdinfo_job(self):
        """:class:`~.jobs.bdinfo.BdinfoJob` instance"""
        return jobs.bdinfo.BdinfoJob(
            precondition=self.make_precondition('bdinfo_job', precondition=self.bdinfo_job_precondition),
            **self.common_job_args(),
        )

    def bdinfo_job_precondition(self):
        return self.is_bdmv_release

    @functools.cached_property
    def video_info(self):
        """
        Map video file paths to mediainfo/bdinfo reports and screenshot URLs

        Every key in the returned :class:`dict` is a file ID, which is derived from the file path
        from which mediainfo reports, bdinfo reports and/or screenshot URLs were generated.

        Every value in the returned :class:`dict` is a :class:`dict` with the keys ``mediainfos``,
        ``bdinfos`` and ``screenshot_urls``. Each value is a sequence, which may be empty.

        ``mediainfos`` is a sequence of mediainfo reports (:class:`str`), ``bdinfos`` is a sequence
        of :class:`~.BdinfoReport` instances, and ``screenshot_urls`` is a sequence of
        :class:`~.UploadedImage` instances.

        .. note:: We allow multiple reports per video file because VIDEO_TS releases may require one
                  report for the .IFO file and another report for the .VOB file.

        Season example:

        .. code::

            {
              "Foo.S01/Foo.S01E01.mkv": {
                "mediainfos": (
                  "<mediainfo report for Foo.S01E01.mkv>",
                ),
                "bdinfos": (),
                "screenshot_urls": (
                  <UploadeImage for Foo.S01E01.mkv.0:12:09.png>,
                  <UploadeImage for Foo.S01E01.mkv.0:24:12.png>,
                ),
              },
              "Foo.S01/Foo.S01E02.mkv": {
                "mediainfos": (
                  "<mediainfo report for Foo.S01E02.mkv>",
                ),
                "bdinfos": (),
                "screenshot_urls": (
                  <UploadedImage for Foo.S01E02.mkv.0:13:12.png>,
                  <UploadedImage for Foo.S01E02.mkv.0:26:03.png>,
                ),
              },
              ...
            }

        VIDEO_TS example:

        .. code::

            {
              "VIDEO_TS/VTS_03": {
                "mediainfos": (
                  "<mediainfo report for VTS_03_0.IFO>",
                  "<mediainfo report for VTS_03_2.VOB>",
                ),
                "bdinfos": (),
                "screenshot_urls": (
                  <UploadedImage for VTS_03_2.VOB.0:14:58.png>,
                  <UploadedImage for VTS_03_2.VOB.0:26:12.png>,
                ),
              },
              "VIDEO_TS/VTS_06": {
                "mediainfos": (
                  "<mediainfo report for VTS_06_0.IFO>",
                  "<mediainfo report for VTS_06_4.VOB>",
                ),
                "bdinfos": (),
                "screenshot_urls": (
                  <UploadedImage for VTS_06_4.VOB.0:12:53.png>,
                  <UploadedImage for VTS_06_4.VOB.0:29:32.png>,
                ),
              },
              ...
            }

        BDMV example:

        .. code::

            {
              "BDMV/STREAM/00003.mpls": {
                "mediainfos": (),
                "bdinfos": (
                  <BdinfoReport for 00003.mpls>,
                ),
                "screenshot_urls": (
                  <UplaodedImage for 00003.mpls.0:11:27.png>,
                  <UploadedImage for 00003.mpls.0:21:42.png>,
                ),
              },
              ...
            }
        """
        docs = collections.defaultdict(lambda: {
            'mediainfos': [],
            'bdinfos': [],
            'screenshot_urls': [],
        })

        @functools.lru_cache
        def get_video_id(video_path):
            # `video_path` is empty if screenshot was precreated by the user.
            if video_path:
                relative_video_path = self.get_relative_file_path(video_path)
                if self.is_video_ts_release:
                    # Remove enough characters from file name so that the return value is the same
                    # for .IFO and .VOB files for the same playlist. This means mediainfo reports
                    # are grouped by playlist, i.e. VTS_xx_0.IFO is ends up together with
                    # VTS_xx_y.VOB.
                    return re.sub(r'\b(VTS_\d+)_\d+\.(?:IFO|VOB)$', r'\1', relative_video_path)
                return relative_video_path
            return video_path

        # Map each created screenshot to its video file and look up the URL(s) for that screenshot.
        # We want any `precreated_screenshots` at the top, so we have to process screenshots first.
        assert self.upload_screenshots_job.is_finished
        urls_by_file = self.upload_screenshots_job.urls_by_file
        screenshots_by_file = self.screenshots_job.screenshots_by_file
        for video_filepath, screenshot_filepaths in screenshots_by_file.items():
            for screenshot_filepath in screenshot_filepaths:
                uploaded_image = urls_by_file.get(screenshot_filepath, None)
                if uploaded_image:
                    docs[get_video_id(video_filepath)]['screenshot_urls'].append(uploaded_image)

        # Map each created mediainfo report to its video file.
        assert self.mediainfo_job.is_finished or not self.mediainfo_job.is_enabled
        for video_filepath, report in self.mediainfo_job.reports_by_file.items():
            docs[get_video_id(video_filepath)]['mediainfos'].append(report)

        # Map each created mediainfo report to its video file.
        assert self.bdinfo_job.is_finished or not self.bdinfo_job.is_enabled
        for video_filepath, report in self.bdinfo_job.reports_by_file.items():
            docs[get_video_id(video_filepath)]['bdinfos'].append(report)

        # Make sequences immutable.
        for doc in docs.values():
            doc['mediainfos'] = tuple(doc['mediainfos'])
            doc['bdinfos'] = tuple(doc['bdinfos'])
            doc['screenshot_urls'] = tuple(doc['screenshot_urls'])

        return dict(docs)

    @functools.cached_property
    def scene_check_job(self):
        """:class:`~.jobs.scene.SceneCheckJob` instance"""
        common_job_args = self.common_job_args(ignore_cache=True)
        common_job_args['force'] = self.options.get('is_scene')
        return jobs.scene.SceneCheckJob(
            content_path=self.content_path,
            precondition=self.make_precondition('scene_check_job'),
            **common_job_args,
        )

    def make_precondition(self, job_attr, precondition=None):
        """
        Return :attr:`~.base.JobBase.precondition` function for job

        The returned function takes into account :attr:`jobs_before_upload`,
        :attr:`jobs_after_upload` and :attr:`isolated_jobs`.

        :param str job_attr: Name of the job attribute this precondition is for

            By convention, this should be ``"<name>_job"``.

        :param callable precondition: Custom :attr:`~.base.JobBase.precondition`

            `precondition` must be either `None` or return anything truthy for
            the job to get enabled.
        """
        def custom_precondition(precondition=precondition):
            return precondition is None or precondition()

        def precondition():
            job = getattr(self, job_attr)
            if not (
                    job in self.jobs_before_upload
                    or job in self.jobs_after_upload
            ):
                # Subclass doesn't use this job
                return False

            isolated_jobs = self.isolated_jobs
            if isolated_jobs and job in isolated_jobs:
                # Jobs was isolated by user (i.e. other jobs are disabled)
                return custom_precondition()

            if not isolated_jobs:
                # No isolated jobs means all jobs in jobs_before/after_upload are enabled
                return custom_precondition()

            return False

        # Rename precondition function to make debugging more readable
        precondition.__qualname__ = f'{job_attr}_precondition'
        return precondition

    _NO_DEFAULT = object()

    def get_job_output(self, job, slice=None, default=_NO_DEFAULT):
        """
        Helper method for getting output from job

        `job` must be finished.

        :param job: :class:`~.jobs.base.JobBase` instance
        :param slice: :class:`int` to get a specific item from `job`'s output,
            `None` to return all output as a list, or a :class:`slice` object to
            return only one or more items of the output
        :param default: Default value if `job` is not finished or getting
            `slice` from `job`'s output fails.

        :raise RuntimeError: if `job` is not finished or getting `slice` from
            :attr:`~.base.JobBase.output` raises an :class:`IndexError`
        :return: :class:`list` or :class:`str`
        """
        if not job.is_finished:
            if default is not self._NO_DEFAULT:
                return default
            else:
                raise RuntimeError(f'Cannot get output from unfinished job: {job.name}')
        else:
            if slice is None:
                slice = builtins.slice(None, None)
            try:
                return job.output[slice]
            except IndexError as e:
                if default is not self._NO_DEFAULT:
                    return default
                else:
                    raise RuntimeError(f'Job finished with insufficient output: {job.name}: {job.output}') from e

    def get_job_attribute(self, job, attribute, default=_NO_DEFAULT):
        """
        Helper method for getting an attribute from job

        :param job: :class:`~.jobs.base.JobBase` instance
        :param str attribute: Name of attribute to get from `job`
        :param default: Default value if `job` is not finished

        :raise RuntimeError: if `job` is not finished
        :raise AttributeError: if `attribute` is not an attribute of `job`
        """
        if not job.is_finished:
            if default is not self._NO_DEFAULT:
                return default
            else:
                raise RuntimeError(f'Cannot get attribute from unfinished job: {job.name}')
        else:
            return getattr(job, attribute)

    def get_relative_file_path(self, file_path):
        """
        Return `file_path` relative to :attr:`content_path`

        The first path component of the returned path is the last component of :attr:`content_path`.

        :raise ValueError: if `file_path` is not a subpath of :attr:`content_path`
        """
        # Make paths absolute and resolve any symbolic links.
        file_path_abs = pathlib.Path(file_path).resolve()
        content_path_abs = pathlib.Path(self.content_path).resolve()

        try:
            subpath = file_path_abs.relative_to(content_path_abs)
        except ValueError as e:
            raise ValueError(f'{str(file_path_abs)!r} is not a subpath of {str(content_path_abs)!r}') from e
        else:
            # Prepend last component of `content_path` (i.e. the release name) to the relative path.
            return str(content_path_abs.name / subpath)

    def make_screenshots_grid(self, screenshots, columns=2, horizontal_spacer='  ', vertical_spacer='\n'):
        """
        Return BBcode for screenshots in a grid layout

        :param screenshots: Sequence of :class:`~.imghosts.common.UploadedImage`
            objects (URLs with a ``thumbnail_url`` attribute)
        :param int columns: How many columns to split screenshots into
        :param str horizontal_spacer: String between columns
        :param str vertical_spacer: String between rows

        :raise RuntimeError: if any screenshot doesn't have a thumbnail
        """
        groups = utils.as_groups(
            screenshots,
            group_sizes=(columns,),
            default=None,
        )

        rows = []
        for screenshots in groups:
            cells = []
            for screenshot in screenshots:
                # `screenshot` is `None` at the end if the number of screenshots
                # is not perfectly divisible by `columns`
                if screenshot is not None:
                    if screenshot.thumbnail_url is None:
                        raise RuntimeError(f'No thumbnail for {screenshot}')
                    cells.append(f'[url={screenshot}][img]{screenshot.thumbnail_url}[/img][/url]')

            # Space between columns
            rows.append(horizontal_spacer.join(cells))

        # Space between rows
        return vertical_spacer.join(rows)

    def read_nfo(self, *, strip=False):
        """
        Return NFO file content from user-supplied file path via :attr:`options`

        If no file path is supplied by the user, find ``*.nfo`` file beneath :attr:`content_path`.

        If no ``*.nfo`` file is found, return `None`.

        If an ``*.nfo`` file is found but cannot be read, call
        :meth:`~.TrackerJobsBase.error`.

        :param path: Path to NFO file or directory that contains an NFO file or `None` to use
            :attr:`content_path`

        See :func:`.string.read_nfo` for more information.
        """
        nfo_filepath = self.options.get('nfo', self.content_path)
        try:
            return utils.string.read_nfo(nfo_filepath, strip=strip)
        except errors.ContentError as e:
            self.error(e)

    @functools.cached_property
    def promotion_bbcode(self):
        """
        Return self promotional BBcode

        If ``only_description`` in :attr:`options` is set, return an empty string.
        """
        if not self.options.get('only_description', False):
            return (
                '[align=right][size=1]'
                f'Shared with [url={__homepage__}]{__project_name__}[/url]'
                '[/size][/align]'
            )
        else:
            return ''



__project_name__ = 'upsies'
__description__ = 'Media metadata aggregator'
__homepage__ = 'https://upsies.readthedocs.io'
__changelog__ = 'https://codeberg.org/plotski/upsies/raw/branch/master/NEWS'
__version__ = '2025.01.02'


def application_setup(config):
    """
    This function should be called by the UI ASAP when `config` is available

    :param config: :class:`~.configfiles.ConfigFiles` instance
    """
    import os

    from . import utils

    utils.http.cache_directory = os.path.join(
        config['config']['main']['cache_directory'],
        'http',
    )


def application_shutdown(config):
    """
    This function should be called by the UI before the applicatin terminates

    :param config: :class:`~.configfiles.ConfigFiles` instance
    """
    import asyncio

    from . import utils

    # Maintain maximum cache size
    utils.fs.limit_directory_size(
        path=config['config']['main']['cache_directory'],
        max_total_size=config['config']['main']['max_cache_size'],
    )

    # Remove empty files and directories
    utils.fs.prune_empty(
        path=config['config']['main']['cache_directory'],
        files=True,
        directories=True,
    )



"""
Concrete :class:`~.base.TrackerJobsBase` subclass for ANT
"""

import functools
import re

from ... import jobs, utils
from ..base import TrackerJobsBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class AntTrackerJobs(TrackerJobsBase):

    @functools.cached_property
    def jobs_before_upload(self):
        return (
            # Interactive jobs
            self.playlists_job,
            self.tmdb_job,
            self.source_job,
            self.scene_check_job,

            # Background jobs
            self.create_torrent_job,
            self.group_job,
            self.mediainfo_job,
            self.bdinfo_job,
            self.flags_job,
            self.anonymous_job,
            self.description_job,
        )

    @functools.cached_property
    def source_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('source'),
            label='Source',
            precondition=self.make_precondition('source_job'),
            options=(
                ('Blu-ray', 'Blu-ray'),
                ('DVD', 'DVD'),
                ('WEB', 'WEB'),
                ('HD-DVD', 'HD-DVD'),
                ('HDTV', 'HDTV'),
                ('VHS', 'VHS'),
                ('TV', 'TV'),
                ('LaserDisc', 'LaserDisc'),
                ('Unknown', 'Unknown'),
            ),
            autodetect=self.autodetect_source,
            autofinish=True,
            **self.common_job_args(),
        )

    _autodetect_source_map = {
        'Blu-ray': lambda release_name: 'BluRay' in release_name.source,
        'HD-DVD': lambda release_name: 'HD-DVD' in release_name.source,
        'HDTV': lambda release_name: 'HDTV' in release_name.source,
        'DVD': lambda release_name: 'DVD' in release_name.source,
        'WEB': lambda release_name: 'WEB' in release_name.source,
        'VHS': lambda release_name: 'VHS' in release_name.source,
        'TV': lambda release_name: 'TV' in release_name.source,
        # 'LaserDisc': lambda release_name: ...,  # Not supported by ReleaseName
    }

    async def autodetect_source(self, job_):
        for option, autodetect in self._autodetect_source_map.items():
            if autodetect(self.release_name):
                return option

    @functools.cached_property
    def group_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('group'),
            label='Group',
            precondition=self.make_precondition('group_job'),
            text=self.autodetect_group,
            validator=self.validate_group,
            finish_on_success=True,
            **self.common_job_args(),
        )

    # When uploading a group with "-" in it, ANT says: "Release group must be alphanumeric only."
    _release_group_regex = re.compile(r'^[a-zA-Z0-9]+$')

    async def autodetect_group(self):
        group = self.release_name.group
        if self._release_group_regex.search(group):
            return group
        else:
            self.group_job.set_text(group)

    def validate_group(self, group):
        if not self._release_group_regex.search(group):
            raise ValueError('Release group must only contain alphanumeric characters (a-z, A-Z, 0-9)')

    @functools.cached_property
    def flags_job(self):
        return jobs.custom.CustomJob(
            name=self.get_job_name('flags'),
            label='Flags',
            precondition=self.make_precondition('flags_job'),
            worker=self.autodetect_flags,
            no_output_is_ok=True,
            **self.common_job_args(ignore_cache=True),
        )

    async def autodetect_flags(self, job):
        # supported flags: Directors, Extended, Uncut, IMAX, Unrated, HDR10, DV,
        # 4KRemaster, Atmos, DualAudio, Commentary, Remux, 3D, Criterion

        flags = []
        rn = self.release_name

        if "Director's Cut" in rn.edition:
            flags.append('Directors')
        if 'Extended Cut' in rn.edition:
            flags.append('Extended')
        if 'Uncut' in rn.edition:
            flags.append('Uncut')
        if 'Unrated' in rn.edition:
            flags.append('Unrated')
        if 'Criterion Collection' in rn.edition:
            flags.append('Criterion')
        if 'IMAX' in rn.edition:
            flags.append('IMAX')
        if '4k Remastered' in rn.edition:
            flags.append('4KRemaster')
        if 'Dual Audio' in rn.edition:
            flags.append('DualAudio')

        if 'Remux' in rn.source:
            flags.append('Remux')

        hdr_formats = utils.mediainfo.video.get_hdr_formats(self.content_path, default=())
        if 'DV' in hdr_formats:
            flags.append('DV')
        if 'HDR10' in hdr_formats or 'HDR10+' in hdr_formats:
            flags.append('HDR10')

        if 'Atmos' in rn.audio_format:
            flags.append('Atmos')

        if self.release_name.has_commentary:
            flags.append('Commentary')

        return flags

    @functools.cached_property
    def anonymous_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('anonymous'),
            label='Anonymous',
            precondition=self.make_precondition('anonymous_job'),
            options=(
                ('No', False),
                ('Yes', True),
            ),
            autodetect=self.autodetect_anonymous,
            autofinish=True,
            **self.common_job_args(ignore_cache=True),
        )

    async def autodetect_anonymous(self, job):
        return self.options.get('anonymous', False)

    @functools.cached_property
    def description_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            precondition=self.make_precondition('description_job'),
            prejobs=(
                self.mediainfo_job,
                self.bdinfo_job,
            ),
            text=self.generate_description,
            hidden=True,
            finish_on_success=True,
            read_only=True,
            **self.common_job_args(ignore_cache=True),
        )

    def generate_description(self):
        parts = []

        # FIXME [2024-08-17] NFOs are currently interpreted as BBcode and/or Markdown even inside
        #                    [pre] tags, wich messes up everything.
        # nfo = self.generate_description_nfo()
        # if nfo:
        #     parts.append(nfo)

        if mediainfo := self.generate_description_mediainfo():
            parts.append(mediainfo)

        if bdinfo := self.generate_description_bdinfo():
            parts.append(bdinfo)

        if parts and self.promotion_bbcode:
            parts.append(f'\n{self.promotion_bbcode}')
        else:
            parts.append(self.promotion_bbcode)

        return '\n'.join(part for part in parts if part)

    def generate_description_bdinfo(self):
        # BDInfo is not generated if we're submitting non-BDMV release.
        if self.bdinfo_job.is_enabled:
            assert self.bdinfo_job.is_finished
            bdinfos_by_file = tuple(self.bdinfo_job.reports_by_file.items())
            if bdinfos_by_file:
                bdinfos_bbcode = []
                for video_filepath, bdinfo in bdinfos_by_file:
                    filetitle = self.get_relative_file_path(video_filepath)
                    bdinfos_bbcode.append(f'[spoiler={filetitle}]{bdinfo.quick_summary}[/spoiler]')
                return '\n'.join(bdinfos_bbcode)

    def generate_description_mediainfo(self):
        # For VIDEO_TS releases, 2 mediainfo reports are generted, one for an .IFO file and one for
        # a .VOB. Because the "mediainfo" in the POST request has only room for one report, we
        # include both reports in the description as well.
        if self.mediainfo_job.is_enabled:
            assert self.mediainfo_job.is_finished
            mediainfos_by_file = tuple(self.mediainfo_job.reports_by_file.items())
            if len(mediainfos_by_file) >= 2:
                mediainfos_bbcode = []
                for video_filepath, mediainfo in mediainfos_by_file:
                    filetitle = self.get_relative_file_path(video_filepath)
                    mediainfos_bbcode.append(f'[spoiler={filetitle}]{mediainfo}[/spoiler]')
                return '\n'.join(mediainfos_bbcode)

    def generate_description_nfo(self):
        nfo = self.read_nfo(strip=True)
        if nfo:
            return (
                '[spoiler=NFO]'
                + '[pre]'
                + nfo
                + '[/pre]'
                + '[/spoiler]'
            )

    # Mediainfo is provided normally and BDInfo is added to the description.
    mediainfo_required_for_bdmv = True

    @property
    def post_data(self):
        return {
            'api_key': self._tracker.apikey,
            'action': 'upload',
            'tmdbid': self.get_job_output(self.tmdb_job, slice=0).replace('movie/', ''),
            'mediainfo': self.get_job_output(self.mediainfo_job, slice=0),
            'release_desc': self.get_job_output(self.description_job, slice=0) or None,
            'flags[]': self.get_job_output(self.flags_job),
            # Scene release? (I don't know why it's called "censored".)
            'censored': '1' if self.get_job_attribute(self.scene_check_job, 'is_scene_release') else None,
            'anonymous': '1' if self.get_job_attribute(self.anonymous_job, 'choice') else None,
            'media': self.get_job_attribute(self.source_job, 'choice'),
            **self._post_data_release_group,
        }

    @property
    def _post_data_release_group(self):
        group = self.get_job_output(self.group_job, slice=0)
        if group != 'NOGROUP':
            return {'releasegroup': group}
        else:
            # Default value of <input type="checkbox"> is "on":
            # https://developer.mozilla.org/en-US/docs/Web/HTML/Element/Input/checkbox
            return {'noreleasegroup': 'on'}

    @property
    def post_files(self):
        return {
            'file_input': {
                'file': self.torrent_filepath,
                'mimetype': 'application/x-bittorrent',
            },
        }



"""
Print ``bdinfo`` output
"""

import functools

from .... import jobs, utils
from .base import CommandBase


class bdinfo(CommandBase):
    """
    Print the BDInfo report(s)

    This needs a ``bdinfo`` executable in your $PATH. It is called like this:

        bdinfo --mpls=PLAYLIST_FILE_NAME PATH_TO_BLURAY REPORT_DIRECTORY

    Here is a wrapper script that runs BDInfoCLI-ng in a docker container:
    https://codeberg.org/plotski/upsies/src/branch/master/goodies/bdinfo
    """

    names = ('bdinfo', 'bi')

    argument_definitions = {
        'CONTENT': {
            'type': utils.argtypes.content,
            'help': 'Path to release content',
        },
        ('--summary', '-s'): {
            'choices': ('none', 'full', 'quick'),
            'default': 'none',
            'help': 'Which parts of the BDInfo output to print',
            'group': 'playlist_or_summary',
        },
        ('--multiple', '-m'): {
            'help': 'Allow selection of multiple playlists and generate multiple BDInfo reports',
            'action': 'store_true',
        },
        ('--format', '-f'): {
            'help': ('Text that contains the placeholder "{BDINFO}", which is replaced '
                     'by the BDInfo report\n'
                     'Example: "[bdinfo]{BDINFO}[/bdinfo]"'),
            'default': '{BDINFO}',
        },
    }

    @functools.cached_property
    def playlists_job(self):
        return jobs.playlists.PlaylistsJob(
            home_directory=self.home_directory,
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
            content_path=self.args.CONTENT,
            select_multiple=self.args.multiple,
        )

    @functools.cached_property
    def bdinfo_job(self):
        return jobs.bdinfo.BdinfoJob(
            home_directory=self.home_directory,
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
            summary=(
                None
                if self.args.summary == 'none' else
                self.args.summary
            ),
            format=self.args.format,
        )

    @functools.cached_property
    def jobs(self):
        return (
            self.playlists_job,
            self.bdinfo_job,
        )
